from abc import ABC  # abstract classes library


class BTree(ABC):
    """An abstract class used to represent a Binary Tree

    ...

    Methods
    -------
    is_leaf()
        Indicates if the BTree is a leaf
    get_value()
        Get the value contained in the leaf
    """

    def is_leaf(self):
        """ Indicates if the BTree is a leaf
        """
        return False

    def is_node(self):
        """ Indicates if the BTree is a node
        """
        return False


class Leaf(BTree):
    """A class that overrides BTree used to represent a Leaf

    ...

    Attributes
    ----------
    value
        A value describing the current leaf

    Methods
    -------
    is_leaf()
        Indicates if the BTree is a leaf
    get_value()
        Get the value contained in the current leaf
    set_value(v)
        Set the value contained in the current leaf
    map(kl, kn)
        Applies functions to every leaf and to every node values
    mapt(kl, kn)
        Applies kl to every leaf values the current instance, and kn to every subtrees that are nodes
    reduce(k)
        Reduces a BTree into a single value using k
    uacc(k)
        Makes an upward accumulation of the values in a BTree using k
    dacc(gl, gr, c)
        Makes an downward accumulation of the values in a BTree using gl, gr and c
    zip(t)
        Zip the values contained in a second BTree with the ones in the current instance
    map2(t, f)
        Zip the values contained in a tree with the ones in the current instance using a function
    getchl(c)
        Shift all the values contained in the current instance by the left
    getchr(c)
        Shift all the values contained in the current instance by the right
    size()
        Gives the number of elements in the current instance
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "leaf " + str(self.value)

    def __eq__(self, other):
        if isinstance(other, Leaf):
            return self.get_value() == other.get_value()
        return False

    def is_leaf(self):
        """Indicates if the BTree is a leaf
        """
        return True

    def get_value(self):
        """Get the value contained in the current leaf
        """
        return self.value

    def set_value(self, v):
        """Set the value contained in the current leaf

        Parameters
        ----------
        v
            The new value to set in the current instance
        """
        self.value = v

    def map(self, kl, kn):
        """Applies functions to every leaf and to every node values

        Parameters
        ----------
        kl : callable
            A function to apply to every leaf values of the current instance
        kn : callable
            A function to apply to every node values of the current instance
        """
        return Leaf(kl(self.get_value()))

    def mapt(self, kl, kn):
        """Applies a function to every leaf values the current instance, and another one to every subtrees that are nodes

        Parameters
        ----------
        kl : callable
            The function to apply to every leaf values of the current instance
        kn : callable
            The function to apply to every node subtrees of the current instance
        """
        return Leaf(kl(self.get_value()))

    def reduce(self, k):
        """Reduces a BTree into a single value using a function k

        If the BTree is a leaf, the single reduced value is the value contained in the structure.

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        """
        return self.get_value()

    def uacc(self, k):
        """Makes an upward accumulation of the values in the current instance using a function k

        If the BTree is a leaf, the tree doesn't change.

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        """
        return Leaf(self.get_value())

    def dacc(self, gl, gr, c):
        """Makes an downward accumulation of the values in a BTree using gl, gr and c

        Parameters
        ----------
        gl : callable
            Function to make an accumulation to the left part of a node
        gr : callable
            Function to make an accumulation to the right part of a node
        c
            Accumulator for the downward computation
        """
        return Leaf(c)

    def zip(self, t):
        """Zip the values contained in t with the ones in the current instance

        Precondition
        -------------
        t should be a Leaf instance

        Parameters
        ----------
        t : :obj:`BTree`
            The BTree to zip with the current instance
        """
        assert t.is_leaf(), "A leaf can only be zipped with another leaf"
        return Leaf((self.get_value(), t.get_value()))

    def map2(self, f, t):
        """Zip the values contained in a tree with the ones in the current instance using a function

        Precondition
        -------------
        t should be a Leaf instance

        Parameters
        ----------
        t : :obj:`BTree`
            The BTree to zip with the current instance
        f : callable
            A function to zip values
        """
        assert t.is_leaf(), "A leaf can only be zipped with another leaf"
        return Leaf(f(self.get_value(), t.get_value()))

    def getchl(self, c):
        """Shift all the values contained in the current instance by the left

        Parameters
        ----------
        c
            The default value for elements that doesn't have left children
        """
        return Leaf(c)

    def getchr(self, c):
        """Shift all the values contained in the current instance by the right

        Parameters
        ----------
        c
            The default value for elements that doesn't have right children
        """
        return Leaf(c)

    def size(self):
        """Gives the number of elements in the current instance
        """
        return 1


class Node(BTree):
    """A class that overrides BTree used to represent a Node

    ...

    Attributes
    ----------
    value
        A value describing the current node
    left : :obj:`BTree`
        The left-subtree of the current node
    right : :obj:`BTree`
        The right-subtree of the current node

    Methods
    -------
    is_node()
        Indicates if the BTree is a node
    get_value()
        Get the value contained in the current node
    set_value(v)
        Set the value contained in the current node
    get_left()
        Get the left subtree of the current node
    get_right()
        Get the right subtree of the current node
    map(kl, kn)
        Applies functions to every leaf and to every node values
    mapt(kl, kn)
        Applies kl to every leaf values the current instance, and kn to every subtrees that are nodes
    reduce(k)
        Reduces a BTree into a single value using a function k
    uacc(k)
        Makes an upward accumulation of the values in a BTree using a function k
    dacc(gl, gr, c)
        Makes an downward accumulation of the values in a BTree using gl, gr and c
    zip(t)
        Zip the values contained in t with the ones in the current instance
    map2(t, f)
        Zip the values contained in a tree with the ones in the current instance using a function
    getchl(c)
        Shift all the values contained in the current instance by the left
    getchr(c)
        Shift all the values contained in the current instance by the right
    size()
        Gives the number of elements in the current instance
    """

    def __init__(self, value, left, right):
        self.value = value
        self.left = left
        self.right = right

    def __str__(self):
        return "node " + str(self.value) + " (" + str(self.left) + ") (" + str(self.right) + ")"

    def __eq__(self, other):
        if isinstance(other, Node):
            return (self.get_value() == other.get_value()) and (self.get_left() == other.get_left()) and (
                        self.get_right() == other.get_right())
        return False

    def is_node(self):
        """Indicates if the BTree is a node
        """
        return True

    def set_value(self, v):
        """Set the value contained in the current node

        Parameters
        ----------
        v
            The new value to set in the current instance
        """
        self.value = v

    def get_value(self):
        """Get the value contained in the current node
        """
        return self.value

    def get_right(self):
        """Get the right subtree of the current node
        """
        return self.right

    def get_left(self):
        """Get the left subtree of the current node
        """
        return self.left

    def map(self, kl, kn):
        """Applies functions to every leaf and to every node values

        Parameters
        ----------
        kl : callable
            A function to apply to every leaf values of the current instance
        kn : callable
            A function to apply to every node values of the current instance
        """
        new_val = kn(self.get_value())
        left = self.get_left().map(kl, kn)
        right = self.get_right().map(kl, kn)
        return Node(new_val, left, right)

    def mapt(self, kl, kn):
        """Applies kl to every leaf values the current instance, and kn to every subtrees that are nodes

        Parameters
        ----------
        kl : callable
            The function to apply to every leaf values of the current instance
        kn : callable
            The function to apply to every node subtrees of the current instance
        """
        new_val = kn(self.get_value(), self.get_left(), self.get_right())
        left = self.get_left().mapt(kl, kn)
        right = self.get_right().mapt(kl, kn)
        return Node(new_val, left, right)

    def reduce(self, k):
        """Reduces a BTree into a single value using a function k

        We use recursive calls of sub-reduction to make a total reduction.

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        """
        left = self.get_left().reduce(k)
        right = self.get_right().reduce(k)
        return k(left, self.get_value(), right)

    def uacc(self, k):
        """Makes an upward accumulation of the values in the current instance using a function k

        Every values in nodes are replaced by the reduced value of the BTree considering the current node as the root.

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        """
        r = self.reduce(k)
        return Node(r, self.get_left().uacc(k), self.get_right().uacc(k))

    def dacc(self, gl, gr, c):
        """Makes an downward accumulation of the values in a BTree using gl, gr and c

        Parameters
        ----------
        gl : callable
            Function to make an accumulation to the left part of a node
        gr : callable
            Function to make an accumulation to the right part of a node
        c
            Accumulator for the downward computation
        """
        b = self.get_value()
        left = self.get_left().dacc(gl, gr, gl(c, b))
        right = self.get_right().dacc(gl, gr, gr(c, b))
        return Node(c, left, right)

    def zip(self, t):
        """Zip the values contained in t with the ones in the current instance

        Precondition
        -------------
        f should be a Node instance

        Parameters
        ----------
        t : :obj:`BTree`
            The BTree to zip with the current instance
        """
        assert t.is_node(), "A node can only be zipped with another node"

        v = (self.get_value(), t.get_value())
        left = self.get_left().zip(t.get_left())
        right = self.get_right().zip(t.get_right())
        return Node(v, left, right)

    def map2(self, f, t):
        """Zip the values contained in a tree with the ones in the current instance using a function

        Precondition
        -------------
        f should be a Node instance

        Parameters
        ----------
        t : :obj:`BTree`
            The BTree to zip with the current instance
        f : callable
            A function to zip values
        """
        assert t.is_node(), "A node can only be zipped with another node"
        v = f(self.get_value(), t.get_value())
        left = self.get_left().map2(f, t.get_left())
        right = self.get_right().map2(f, t.get_right())
        return Node(v, left, right)

    def getchl(self, c):
        """Shift all the values contained in the current instance by the left

        Parameters
        ----------
        c
            The default value for elements that doesn't have left children
        """
        v = self.get_left().get_value()
        left = self.get_left().getchl(c)
        right = self.get_right().getchl(c)
        return Node(v, left, right)

    def getchr(self, c):
        """Shift all the values contained in the current instance by the right

        Parameters
        ----------
        c
            The default value for elements that doesn't have right children
        """
        v = self.get_right().get_value()
        left = self.get_left().getchr(c)
        right = self.get_right().getchr(c)
        return Node(v, left, right)

    def size(self):
        """Gives the number of elements in the current instance
        """
        return 1 + self.get_left().size() + self.get_right().size()



from pyske.core.tree.ltree import LTree, Segment, TaggedValue
from pyske.core.tree.btree import Node, Leaf

# -------------------------- #

def test_serialization_1_4():
	#btree4
	bt = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	m = 4
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(1,"C")])
	seg2 = Segment([TaggedValue(2,"C")])
	seg3 = Segment([TaggedValue(4,"N"),TaggedValue(6,"L"),TaggedValue(7,"L")])
	seg4 = Segment([TaggedValue(5,"N"),TaggedValue(8,"L"),TaggedValue(9,"L")])
	seg5 = Segment([TaggedValue(3,"L")])
	exp = LTree([seg1, seg2, seg3, seg4, seg5])
	assert res == exp


def test_serialization_1_3():
	#btree4
	bt = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	m = 3
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(1,"N"), TaggedValue(2,"C"), TaggedValue(3,"L")])
	seg2 = Segment([TaggedValue(4,"N"),TaggedValue(6,"L"),TaggedValue(7,"L")])
	seg3 = Segment([TaggedValue(5,"N"),TaggedValue(8,"L"),TaggedValue(9,"L")])
	exp = LTree([seg1, seg2, seg3])
	assert res == exp



def test_serialization_1_2():
	#btree4
	bt = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	m = 2
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(1,"C")])
	seg2 = Segment([TaggedValue(2,"C")])
	seg3 = Segment([TaggedValue(4,"C")])
	seg4 = Segment([TaggedValue(6,"L")])
	seg5 = Segment([TaggedValue(7,"L")])
	seg6 = Segment([TaggedValue(5,"C")])
	seg7 = Segment([TaggedValue(8,"L")])
	seg8 = Segment([TaggedValue(9,"L")])
	seg9 = Segment([TaggedValue(3,"L")])
	exp = LTree([seg1, seg2, seg3, seg4, seg5, seg6, seg7, seg8, seg9])
	assert res == exp


def test_serialization_1_1():
	#btree4
	bt = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	m = 1
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(1,"C")])
	seg2 = Segment([TaggedValue(2,"C")])
	seg3 = Segment([TaggedValue(4,"C")])
	seg4 = Segment([TaggedValue(6,"L")])
	seg5 = Segment([TaggedValue(7,"L")])
	seg6 = Segment([TaggedValue(5,"C")])
	seg7 = Segment([TaggedValue(8,"L")])
	seg8 = Segment([TaggedValue(9,"L")])
	seg9 = Segment([TaggedValue(3,"L")])
	exp = LTree([seg1, seg2, seg3, seg4, seg5, seg6, seg7, seg8, seg9])
	assert res == exp


def test_serialization_2_5():
	#btree5
	bt =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	m = 5
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg3 = Segment([TaggedValue(9,"N"),TaggedValue(7,"C"),TaggedValue(1,"L")])
	seg4 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg5 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	exp = LTree([seg1,seg2,seg3,seg4,seg5])
	assert res == exp


def test_serialization_2_4():
	#btree5
	bt =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	m = 4
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg3 = Segment([TaggedValue(9,"C")])
	seg4 = Segment([TaggedValue(7,"C")])
	seg5 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg6 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg7 = Segment([TaggedValue(1,"L")])
	exp = LTree([seg1,seg2,seg3,seg4,seg5,seg6,seg7])
	assert res == exp


def test_serialization_2_3():
	#btree5
	bt =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	m = 3
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg3 = Segment([TaggedValue(9,"N"),TaggedValue(7,"C"),TaggedValue(1,"L")])
	seg4 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg5 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	exp = LTree([seg1,seg2,seg3,seg4,seg5])
	assert res == exp


def test_serialization_2_2():
	#btree5
	bt =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	m = 2
	res = LTree.init_from_bt(bt, m)
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"C")])
	seg3 = Segment([TaggedValue(1,"L")])
	seg4 = Segment([TaggedValue(1,"L")])
	seg5 = Segment([TaggedValue(9,"C")])
	seg6 = Segment([TaggedValue(7,"C")])
	seg7 = Segment([TaggedValue(3,"C")])
	seg8 = Segment([TaggedValue(1,"L")])
	seg9 = Segment([TaggedValue(1,"L")])
	seg10 = Segment([TaggedValue(3,"C")])
	seg11 = Segment([TaggedValue(1,"L")])
	seg12 = Segment([TaggedValue(1,"L")])
	seg13 = Segment([TaggedValue(1,"L")])
	exp = LTree([seg1, seg2, seg3, seg4, seg5, seg6, seg7, seg8, seg9, seg10, seg11, seg12, seg13])
	assert res == exp

# -------------------------- #

def test_deserialization_1_4():
	#btree4
	seg1 = Segment([TaggedValue(1,"C")])
	seg2 = Segment([TaggedValue(2,"C")])
	seg3 = Segment([TaggedValue(4,"N"),TaggedValue(6,"L"),TaggedValue(7,"L")])
	seg4 = Segment([TaggedValue(5,"N"),TaggedValue(8,"L"),TaggedValue(9,"L")])
	seg5 = Segment([TaggedValue(3,"L")])
	lt = LTree([seg1, seg2, seg3, seg4, seg5])
	res = lt.deserialization()
	exp = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	assert res == exp


def test_deserialization_1_3():
	#btree4
	exp = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	seg1 = Segment([TaggedValue(1,"N"), TaggedValue(2,"C"), TaggedValue(3,"L")])
	seg2 = Segment([TaggedValue(4,"N"),TaggedValue(6,"L"),TaggedValue(7,"L")])
	seg3 = Segment([TaggedValue(5,"N"),TaggedValue(8,"L"),TaggedValue(9,"L")])
	lt = LTree([seg1, seg2, seg3])
	res = lt.deserialization()
	assert res == exp



def test_deserialization_1_2():
	#btree4
	exp = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	seg1 = Segment([TaggedValue(1,"C")])
	seg2 = Segment([TaggedValue(2,"C")])
	seg3 = Segment([TaggedValue(4,"C")])
	seg4 = Segment([TaggedValue(6,"L")])
	seg5 = Segment([TaggedValue(7,"L")])
	seg6 = Segment([TaggedValue(5,"C")])
	seg7 = Segment([TaggedValue(8,"L")])
	seg8 = Segment([TaggedValue(9,"L")])
	seg9 = Segment([TaggedValue(3,"L")])
	lt = LTree([seg1, seg2, seg3, seg4, seg5, seg6, seg7, seg8, seg9])
	res = lt.deserialization()
	assert res == exp


def test_deserialization_1_1():
	#btree4
	exp = Node(1, Node(2, Node (4, Leaf (6), Leaf(7)), Node (5, Leaf (8), Leaf(9))), Leaf(3))
	seg1 = Segment([TaggedValue(1,"C")])
	seg2 = Segment([TaggedValue(2,"C")])
	seg3 = Segment([TaggedValue(4,"C")])
	seg4 = Segment([TaggedValue(6,"L")])
	seg5 = Segment([TaggedValue(7,"L")])
	seg6 = Segment([TaggedValue(5,"C")])
	seg7 = Segment([TaggedValue(8,"L")])
	seg8 = Segment([TaggedValue(9,"L")])
	seg9 = Segment([TaggedValue(3,"L")])
	lt = LTree([seg1, seg2, seg3, seg4, seg5, seg6, seg7, seg8, seg9])
	res = lt.deserialization()
	assert res == exp


def test_deserialization_2_5():
	#btree5
	exp =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg3 = Segment([TaggedValue(9,"N"),TaggedValue(7,"C"),TaggedValue(1,"L")])
	seg4 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg5 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	lt = LTree([seg1,seg2,seg3,seg4,seg5])
	res = lt.deserialization()
	assert res == exp


def test_deserialization_2_4():
	#btree5
	exp =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg3 = Segment([TaggedValue(9,"C")])
	seg4 = Segment([TaggedValue(7,"C")])
	seg5 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg6 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg7 = Segment([TaggedValue(1,"L")])
	lt = LTree([seg1,seg2,seg3,seg4,seg5,seg6,seg7])
	res = lt.deserialization()
	assert res == exp


def test_deserialization_2_3():
	#btree5
	exp =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg3 = Segment([TaggedValue(9,"N"),TaggedValue(7,"C"),TaggedValue(1,"L")])
	seg4 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	seg5 = Segment([TaggedValue(3,"N"),TaggedValue(1,"L"),TaggedValue(1,"L")])
	lt = LTree([seg1,seg2,seg3,seg4,seg5])
	res = lt.deserialization()
	assert res == exp


def test_deserialization_2_2():
	#btree5
	exp =  Node(13, Node(3, Leaf(1), Leaf(1)), Node(9, Node(7,Node(3, Leaf(1), Leaf(1)),Node(3, Leaf(1), Leaf(1))), Leaf(1)))
	seg1 = Segment([TaggedValue(13,"C")])
	seg2 = Segment([TaggedValue(3,"C")])
	seg3 = Segment([TaggedValue(1,"L")])
	seg4 = Segment([TaggedValue(1,"L")])
	seg5 = Segment([TaggedValue(9,"C")])
	seg6 = Segment([TaggedValue(7,"C")])
	seg7 = Segment([TaggedValue(3,"C")])
	seg8 = Segment([TaggedValue(1,"L")])
	seg9 = Segment([TaggedValue(1,"L")])
	seg10 = Segment([TaggedValue(3,"C")])
	seg11 = Segment([TaggedValue(1,"L")])
	seg12 = Segment([TaggedValue(1,"L")])
	seg13 = Segment([TaggedValue(1,"L")])
	lt = LTree([seg1, seg2, seg3, seg4, seg5, seg6, seg7, seg8, seg9, seg10, seg11, seg12, seg13])
	res = lt.deserialization()
	assert res == exp



from pyske.core.list.slist import SList
from pyske.core.list.plist import PList
from pyske.core.support.parallel import *
import sys
import random
from mpi4py import MPI

# Generating a parallel list of the size specified on the command line or 1000
if len(sys.argv) > 1:
    size = int(sys.argv[1])
else:
    size = 1000
X = PList.init(lambda _: random.randint(0, 100), size)

comm.barrier()

# start timing
t = PList.init(lambda _: wtime(), nprocs)

# computing the variance
add = lambda x, y: x + y
n = X.length()
avg = X.reduce(add) / n
var = (X.map(lambda x: (x - avg) ** 2).reduce(add) ** 0.5) / n

# stop timing
elapsed = t.map(lambda x: wtime() - x)
max_elapsed = elapsed.reduce(max)
avg_elapsed = elapsed.reduce(add) / nprocs
all_elapsed = elapsed.mapi(lambda i, x: "[" + str(i) + "]:" + str(x)).to_seq()

# output at processor 0
at_root(lambda:
        print(f'Variance:\t{var}\nTime (max):\t{max_elapsed}\n'
              f'Time (avg):\t{avg_elapsed}\nTime (all):\t{all_elapsed}'))



import logging

from pyske.core.tree.ltree import TaggedValue, Segment, LTree
from pyske.core.support.parallel import *
from pyske.core.support.separate import *


TAG_BASE = "270995"
TAG_COMM_REDUCE = int(TAG_BASE + "11")
TAG_COMM_UACC_1 = int(TAG_BASE + "21")
TAG_COMM_UACC_2 = int(TAG_BASE + "22")
TAG_COMM_DACC_1 = int(TAG_BASE + "31")
TAG_COMM_DACC_2 = int(TAG_BASE + "32")
TAG_COMM_MERGE = int(TAG_BASE + "00")
TAG_TO_SEQ = int(TAG_BASE + "01")

logging.basicConfig(filename='run_ptree.log', level=logging.DEBUG)
with open('run_ptree.log', 'w'):
    pass
logger = logging.getLogger('ptree')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(funcName)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

class PTree:
    """A class used to represent a distributed tree

    Attributes
    ----------
    __distribution: pid -> nb of segments
    __global_index: num seg -> (start, offset)
    __start_index: index of first index for the current pid in global_index
    __nb_segs: nb of indexes for the current pid in global_index
    __content: concatenation of the segments contained in the current instance
    """

    def __init__(self, lt=None):
        self.__distribution = SList([])
        self.__global_index = SList([])
        self.__start_index = 0
        self.__nb_segs = 0
        self.__content = SList([])
        if lt is not None:
            (distribution, global_index) = distribute_tree(lt, nprocs)
            self.__distribution = distribution
            self.__global_index = global_index
            self.__start_index = distribution.scanl(lambda x, y: x + y, 0)[pid]
            self.__nb_segs = distribution[pid]
            for i_seg in range(self.__start_index, self.__start_index + self.__nb_segs):
                self.__content.extend(lt[i_seg])

    def __eq__(self, other):
        if isinstance(other, PTree):
            return self.__distribution == other.__distribution \
                   and self.__global_index == other.__distribution \
                   and self.__start_index == other.__start_index \
                   and self.__nb_segs == other.__nb_segs \
                   and self.__content == other.__content
        return False

    @staticmethod
    def init(pt, content):
        """Factory for distributed tree.

        Creates a PTree from a already existing one, but changing its content

        Parameters
        ----------
        pt : :obj:`PTree`
            A parallel tree that we want to copy the distribution
        content : :obj:`SList`
            The content of the resulting PTree
        """
        p = PTree()
        p.__distribution = pt.__distribution
        p.__global_index = pt.__global_index
        p.__start_index = pt.__start_index
        p.__nb_segs = pt.__nb_segs
        p.__content = content
        return p

    @staticmethod
    def init_from_file(filename, parser=int):
        """Instantiate a distributed tree from a file

        Parameters
        ----------
        filename : str
            The name of the file that contains the PTree to instantiate
        parser : callable, optional
            A function that transforms a string into a specific type.
            By default, string to int
        """
        filename = filename + "." + str(pid)

        def __parser_couple(s):
            s = s.replace("(", "")
            s = s.replace(")", "")
            ss = s.split(",")
            return int(ss[0]), int(ss[1])

        p = PTree()
        content = SList([])
        with open(filename, "r") as f:
            count_line = 0
            for line in f:
                if line.strip()[0] == '#':
                    continue
                # __distribution: pid -> nb of segments
                # __global_index: num seg -> (start, offset)
                if count_line == 0:  # Get the distribution
                    p.__distribution = SList.from_str(line)
                    p.__start_index = p.__distribution.scanl(lambda x, y: x + y, 0)[pid]
                    p.__nb_segs = p.__distribution[pid]
                elif count_line == 1:  # Get the global_index
                    p.__global_index = SList.from_str(line, parser=__parser_couple)
                else:  # Get the content
                    content.extend(Segment.from_str(line, parser=parser))
                count_line = count_line + 1
        p.__content = content
        return p

    def __str__(self):
        return "pid[" + str(pid) + "]:\n" + \
               "  global_index: " + str(self.__global_index) + "\n" + \
               "  distribution: " + str(self.__distribution) + "\n" + \
               "  nb_segs: " + str(self.__nb_segs) + "\n" + \
               "  start_index: " + str(self.__start_index) + "\n" + \
               "  content: " + str(self.__content) + "\n"

    def browse(self):
        """Browse the linearized distributed tree contained in the current processor
        """
        res = "pid[" + str(pid) + "] "
        for (start, offset) in self.__global_index[self.__start_index: self.__start_index + self.__nb_segs]:
            seg = Segment(self.__content[start:start + offset])
            res = res + "\n   " + str(seg)
        return res

    def map(self, kl, kn):
        """Map skeleton for distributed tree

        Parameters
        ----------
        kl : callable
            Function to apply to every leaf value of the current instance
        kn : callable
            Function to apply to every node value of the current instance
        """
        content = SList([None] * self.__content.length())
        logger.debug(
            '[START] pid[' + str(pid) + '] map skeleton')
        for (start, offset) in self.__global_index[self.__start_index: self.__start_index + self.__nb_segs]:
            logger.debug('[START] pid['+str(pid)+'] map_local from ' + str(start) + ' to '+ str(start + offset))
            content[start:start + offset] = Segment(self.__content[start:start + offset]).map_local(kl, kn)
            logger.debug('[END] pid['+str(pid)+'] map_local from ' + str(start) + ' to '+ str(start + offset))
        res = PTree.init(self, content)
        logger.debug(
            '[END] pid[' + str(pid) + '] map skeleton')
        return res

    def reduce(self, k, phi, psi_n, psi_l, psi_r):
        """Reduce skeleton for distributed tree

        The parameters must respect these equalities (closure property):
        * k(l, b, r) = psi_n(l, phi(b), r)
        * psi_n(psi_n(x, l, y), b, r) = psi_n(x, psi_l(l,b,r), y)
        * psi_n(l, b, psi_n(x, r, y)) = psi_n(x, psi_r(l,b,r), y)

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        phi : callable
            A function used to respect the closure property to allow partial computation
        psi_n : callable
            A function used to respect the closure property to make partial computation
        psi_l : callable
            A function used to respect the closure property to make partial computation on the left
        psi_r : callable
            A function used to respect the closure property to make partial computation on the right
        """
        logger.debug(
            '[START] pid[' + str(pid) + '] reduce skeleton')
        # Step 1 : Local Reduction
        gt = Segment([None] * self.__nb_segs)
        i = 0
        for (start, offset) in self.__global_index[self.__start_index: self.__start_index + self.__nb_segs]:
            logger.debug('[START] pid['+str(pid)+'] reduce_local from ' + str(start) + ' to '+ str(start + offset))
            gt[i] = Segment(self.__content[start:start + offset]).reduce_local(k, phi, psi_l, psi_r)
            logger.debug('[END] pid['+str(pid)+'] reduce_local from ' + str(start) + ' to '+ str(start + offset))
            i = i+1
        # Step 2 : Gather local Results
        if pid == 0:
            for i in range(1, nprocs):
                logger.debug(
                    '[START] pid[' + str(pid) + '] reception from ' + str(i))
                gt.extend(comm.recv(source=i, tag=TAG_COMM_REDUCE)['c'])
                logger.debug(
                    '[END] pid[' + str(pid) + '] reception from ' + str(i))
        else:
            logger.debug(
                    '[START] pid[' + str(pid) + '] emission to ' + str(0))
            comm.send({'c': gt}, dest=0, tag=TAG_COMM_REDUCE)
            logger.debug(
                    '[END] pid[' + str(pid) + '] emission to ' + str(0))
        # Step 3 : Global Reduction
        at_root(lambda: logger.debug('[START] pid[' + str(pid) + '] reduce_global'))
        res = gt.reduce_global(psi_n) if pid == 0 else None
        at_root(lambda: logger.debug('[END] pid[' + str(pid) + '] reduce_global'))
        logger.debug(
            '[END] pid[' + str(pid) + '] reduce skeleton')
        return res

    def uacc(self, k, phi, psi_n, psi_l, psi_r):
        """Upward accumulation skeleton for distributed tree

        The parameters must respect these equalities (closure property):
        * k(l, b, r) = psi_n(l, phi(b), r)
        * psi_n(psi_n(x, l, y), b, r) = psi_n(x, psi_l(l,b,r), y)
        * psi_n(l, b, psi_n(x, r, y)) = psi_n(x, psi_r(l,b,r), y)

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        phi : callable
            A function used to respect the closure property to allow partial computation
        psi_n : callable
            A function used to respect the closure property to make partial computation
        psi_l : callable
            A function used to respect the closure property to make partial computation on the left
        psi_r : callable
            A function used to respect the closure property to make partial computation on the right
        """
        logger.debug(
            '[START] pid[' + str(pid) + '] uAcc skeleton')
        assert self.__distribution != []
        # Step 1 : Local Upwards Accumulation
        gt = Segment([None] * self.__nb_segs)
        lt2 = SList([None] * self.__nb_segs)
        i = 0
        for (start, offset) in self.__global_index[self.__start_index: self.__start_index + self.__nb_segs]:
            logger.debug(
                '[START] pid[' + str(pid) + '] uacc_local from ' + str(start) + ' to ' + str(start + offset))
            (top, res) = Segment(self.__content[start:start + offset]).uacc_local(k, phi, psi_l, psi_r)
            logger.debug(
                '[END] pid[' + str(pid) + '] uacc_local from ' + str(start) + ' to ' + str(start + offset))
            gt[i] = top
            lt2[i] = res
            i = i + 1

        # Step 2 : Gather local Results
        if pid == 0:
            for iproc in range(1, nprocs):
                logger.debug(
                    '[START] pid[' + str(pid) + '] reception local from ' + str(i))
                gt.extend(comm.recv(source=iproc, tag=TAG_COMM_UACC_1)['c'])
                logger.debug(
                    '[END] pid[' + str(pid) + '] reception local from ' + str(i))
        else:
            logger.debug(
                '[START] pid[' + str(pid) + '] emission local to ' + str(0))
            comm.send({'c': gt}, dest=0, tag=TAG_COMM_UACC_1)
            logger.debug(
                '[END] pid[' + str(pid) + '] emission local to ' + str(0))

        # Step 3 : Global Upward Accumulation
        gt2 = None
        if pid == 0:
            at_root(lambda: logger.debug('[START] pid[' + str(pid) + '] uacc_global'))
            gt2 = gt.uacc_global(psi_n)
            for i in range(len(gt2)):
                if gt2[i].is_node():
                    gt2[i] = TaggedValue((gt2.get_left(i).get_value(), gt2.get_right(i).get_value()), gt2[i].get_tag())
            at_root(lambda: logger.debug('[END] pid[' + str(pid) + '] uacc_global'))

        # Step 4 : Distributing Global Result
        start = 0
        if pid == 0:
            for iproc in range(nprocs):
                iproc_off = self.__distribution[iproc]
                if iproc != 0:
                    logger.debug(
                        '[START] pid[' + str(pid) + '] emission global to ' + str(iproc))
                    comm.send({'g': gt2[start: start + iproc_off]}, dest=iproc, tag=TAG_COMM_UACC_2)
                    logger.debug(
                        '[END] pid[' + str(pid) + '] emission global to ' + str(iproc))
                start = start + iproc_off
        else:
            logger.debug(
                '[START] pid[' + str(pid) + '] reception global from ' + str(0))
            gt2 = comm.recv(source=0, tag=TAG_COMM_UACC_2)['g']
            logger.debug(
                '[END] pid[' + str(pid) + '] reception global from ' + str(0))

        # Step 5 : Local Updates
        content = SList([None] * self.__content.length())
        for i in range(len(self.__global_index[self.__start_index: self.__start_index + self.__nb_segs])):
            (start, offset) = self.__global_index[self.__start_index: self.__start_index + self.__nb_segs][i]
            logger.debug('[START] pid['+str(pid)+'] uacc_update from ' + str(start) + ' to '+ str(start + offset))
            if gt[i].is_node():
                (lc, rc) = gt2[i].get_value()
                val = Segment(self.__content[start:start + offset]).uacc_update(lt2[i], k, lc, rc)
            else:
                val = lt2[i]
            logger.debug('[END] pid['+str(pid)+'] uacc_update from ' + str(start) + ' to '+ str(start + offset))
            content[start:start + offset] = val
        res = PTree.init(self, content)
        logger.debug(
            '[END] pid[' + str(pid) + '] uAcc skeleton')
        return res

    def dacc(self, gl, gr, c, phi_l, phi_r, psi_u, psi_d):
        """Downward accumulation skeleton for distributed tree

        The parameters must respect these equalities (closure property):
        * gl(c, b) = psi_d(c, phi_l(b))
        * gr(c, b) = psi_d(c, phi_r(b))
        * psi_d(psi_d(c, b), a) = psi_d(c, psi_u(b,a))

        Parameters
        ---------
        gl : callable
            The function used to make an accumulation to the left children in a binary tree
        gr : callable
            The function used to make an accumulation to the right children in a binary tree
        c
            Initial value of accumulation
        phi_l : callable
            A function used to respect the closure property to allow partial computation on the left
        phi_r : callable
            A function used to respect the closure property to allow partial computation on the right
        psi_d : callable
            A function used to respect the closure property to make partial downward accumulation
        psi_u : callable
            A function used to respect the closure property to make partial computation
        """
        logger.debug(
            '[START] pid[' + str(pid) + '] dAcc skeleton')
        # Step 1 : Computing Local Intermediate Values
        gt = Segment([None] * self.__nb_segs)
        i = 0
        for (start, offset) in self.__global_index[self.__start_index: self.__start_index + self.__nb_segs]:
            seg = Segment(self.__content[start:start + offset])
            logger.debug(
                '[START] pid[' + str(pid) + '] dacc_path from ' + str(start) + ' to ' + str(start + offset))
            if seg.has_critical():
                gt[i] = seg.dacc_path(phi_l, phi_r, psi_u)
            else:
                gt[i] = TaggedValue(seg[0].get_value(), "L")
            logger.debug(
                '[END] pid[' + str(pid) + '] dacc_path from ' + str(start) + ' to ' + str(start + offset))
            i = i + 1
        # Step 2 : Gather Local Results
        if pid == 0:
            for iproc in range(1, nprocs):
                logger.debug(
                    '[START] pid[' + str(pid) + '] reception update from ' + str(i))
                gt.extend(comm.recv(source=iproc, tag=TAG_COMM_DACC_1)['c'])
                logger.debug(
                    '[END] pid[' + str(pid) + '] reception update from ' + str(i))
        else:
            logger.debug(
                '[START] pid[' + str(pid) + '] emission update to ' + str(0))
            comm.send({'c': gt}, dest=0, tag=TAG_COMM_DACC_1)
            logger.debug(
                '[END] pid[' + str(pid) + '] emission update to ' + str(0))
        # Step 3 : Global Downward Accumulation
        at_root(lambda: logger.debug('[START] pid[' + str(pid) + '] dacc_global'))
        gt2 = (gt.dacc_global(psi_d, c) if pid == 0 else None)
        at_root(lambda: logger.debug('[END] pid[' + str(pid) + '] dacc_global'))
        # Step 4 : Distributing Global Result
        if pid == 0:
            start = 0
            for iproc in range(nprocs):
                iproc_off = self.__distribution[iproc]
                if iproc != 0:
                    logger.debug(
                        '[START] pid[' + str(pid) + '] emission global to ' + str(iproc))
                    comm.send({'g': gt2[start: start + iproc_off]}, dest=iproc, tag=TAG_COMM_DACC_2)
                    logger.debug(
                        '[END] pid[' + str(pid) + '] emission global to ' + str(iproc))
                start = start + iproc_off
        else:
            logger.debug(
                '[START] pid[' + str(pid) + '] reception global from ' + str(0))
            gt2 = comm.recv(source=0, tag=TAG_COMM_DACC_2)['g']
            logger.debug(
                '[END] pid[' + str(pid) + '] reception global from ' + str(0))
        # Step 5 : Local Downward Accumulation
        content = SList([None] * self.__content.length())
        for i in range(len(self.__global_index[self.__start_index: self.__start_index + self.__nb_segs])):
            (start, offset) = self.__global_index[self.__start_index: self.__start_index + self.__nb_segs][i]
            logger.debug(
                '[START] pid[' + str(pid) + '] dacc_local from ' + str(start) + ' to ' + str(start + offset))
            content[start:start + offset] = Segment(self.__content[start:start + offset]).dacc_local(gl, gr, gt2[i].get_value())
            logger.debug(
                '[END] pid[' + str(pid) + '] dacc_local from ' + str(start) + ' to ' + str(start + offset))
        res = PTree.init(self, content)
        logger.debug(
            '[END] pid[' + str(pid) + '] dAcc skeleton')
        return res

    def zip(self, pt):
        """Zip skeleton for distributed tree

        Precondition
        -------------
        The distributions of self and pt should be the same

        Parameters
        ----------
        pt : :obj:`PTree`
            The PTree to zip with the current instance
        """
        logger.debug(
            '[START] pid[' + str(pid) + '] zip skeleton')
        assert self.__distribution == pt.__distribution
        content = SList([None] * self.__content.length())
        for i in range(len(self.__global_index[self.__start_index: self.__start_index + self.__nb_segs])):
            (start, offset) = self.__global_index[self.__start_index: self.__start_index + self.__nb_segs][i]
            logger.debug('[START] pid[' + str(pid) + '] zip_local from ' + str(start) + ' to ' + str(start + offset))
            content[start:start + offset] = Segment(self.__content[start:start+offset]).zip(Segment(pt.__content[start:start+offset]))
            logger.debug('[END] pid['+ str(pid) + '] zip_local from ' + str(start) + ' to ' + str(start + offset))
        res = PTree.init(self, content)
        logger.debug(
            '[END] pid[' + str(pid) + '] zip skeleton')
        return res

    def map2(self, f, pt):
        """Map2 skeleton for distributed tree

        Precondition
        -------------
        The distributions of self and pt should be the same

        Parameters
        ----------
        pt : :obj:`LTree`
            The LTree to zip with the current instance
        f : callable
            A function to zip values
        """
        logger.debug(
            '[START] pid[' + str(pid) + '] map2 skeleton')
        assert self.__distribution == pt.__distribution
        content = SList([None] * self.__content.length())
        for i in range(len(self.__global_index[self.__start_index: self.__start_index + self.__nb_segs])):
            (start, offset) = self.__global_index[self.__start_index: self.__start_index + self.__nb_segs][i]
            logger.debug('[START] pid[' + str(pid) + '] map2_local from ' + str(start) + ' to ' + str(start + offset))
            content[start:start + offset] = Segment(self.__content[start:start + offset]).map2(f, Segment(
                pt.__content[start:start + offset]))
            logger.debug('[END] pid[' + str(pid) + '] map2_local from ' + str(start) + ' to ' + str(start + offset))
        res = PTree.init(self, content)
        logger.debug(
            '[END] pid[' + str(pid) + '] map2 skeleton')
        return res

    def get_full_index(self):
        def f(x, y):
            (x1, y1) = x
            (x2, y2) = y
            return x1 + y1, y2
        return SList(self.__global_index.scanr(f))

    def to_seq(self):
        full_content = []
        if pid == 0:
            full_index = self.get_full_index()
            res = LTree([None] * full_index.length())
            full_content.extend(self.__content)
            for iproc in range(1, nprocs):
                full_content.extend(comm.recv(source=iproc, tag=TAG_TO_SEQ)['c'])

            for i in range(full_index.length()):
                (start, offset) = full_index[i]
                res[i] = full_content[start:start + offset]
            return res
        else:
            comm.send({'c': self.__content}, dest=0, tag=TAG_TO_SEQ)
            return None



from pyske.core.tree.ltree import LTree
from pyske.core.support.separate import *
import argparse

def create_pt_files(lt, n, filename):
	(distribution, global_index) = distribute_tree(lt, n)
	passed_seg = 0
	for pid in range(n):
		nb_segs = distribution[pid]
		filename_pid = filename + "." + str(pid)
		content_pid = ""
		for i_seg in range(passed_seg, passed_seg+ nb_segs):
			content_pid = content_pid + str(lt[i_seg]) + "\n"
		with open(filename_pid,"w+") as f: 
			f.write(str(distribution) + "\n" + str(global_index) + "\n" + content_pid)
		f.close()
		passed_seg += nb_segs


parser = argparse.ArgumentParser()
parser.add_argument("-i", help="name of the file that contains the LTree to split")
parser.add_argument("-o", help="name of the files that will contained the PTree (split)")
parser.add_argument("-n", help="how many files we want to get", type=int)
args = parser.parse_args()

filename_in = args.i
filename_out = args.o
n = args.n
if filename_in[-3:] != ".lt":
	filename_in = filename_in + ".lt"

if filename_out[-3:] != ".pt":
	filename_out = filename_out + ".pt"

lt = LTree.init_from_file(filename_in)
create_pt_files(lt, n, filename_out)


class NotEqualSizeError(Exception):
    """
    Exception raised for different sizes between structures

    Attributes
    ----------
    message:
        explanation of the error
    """

    def __init__(self, message):
        self.message = "[NotEqualSizeError] " + message


class EmptyError(Exception):
    """
    Exception raised if a function is applied to an empty structure

    Attributes
    ----------
    message:
        explanation of the error
    """

    def __init__(self, message):
        self.message = "[EmptyError] " + message


class UnknownTypeError(Exception):
    """
    Exception raised if a type is unknown

    Attributes
    ----------
    message:
        explanation of the error
    """

    def __init__(self, message):
        self.message = "[UnknownTypeError] " + message


class IllFormedError(Exception):
    """
    Exception raised if a function is applied to a ill formed structure

    Attributes
    ----------
    message:
        explanation of the error
    """

    def __init__(self, message):
        self.message = "[IllFormedError] " + message


class ApplicationError(Exception):
    """
    Exception raised if a function is applied to not the good structure

    Attributes
    ----------
    message: 
        explanation of the error
    """
    
    def __init__(self, message):
        self.message = "[ApplicationError] " + message


class NotSameTagError(Exception):
    """
    Exception raised if a function is applied to several tagged value with not the same tag

    Attributes
    ----------
    message: 
        explanation of the error
    """

    def __init__(self, message):
        self.message = "[EmptyError] " + message


class ConstructorError(Exception):
    """
    Exception raised if an instance cannot be created

    Attributes
    ----------
    message: 
        explanation of the error
    """
    
    def __init__(self, message):
        self.message = "[ConstructorError] " + message



from pyske.core.support.generate import generate_balanced_btree, generate_random_btree, generate_illbalanced_btree
from pyske.core.tree.ltree import LTree
import random, argparse, math

# --------------------- #

parser = argparse.ArgumentParser()
parser.add_argument("-f", help="name of the file to create to store the BTree")
parser.add_argument("-s", help="size of BTree to generate", type=int)
parser.add_argument("-t", help="type of BTree to generate")
parser.add_argument("-m", help="m value for serialization", type=int)
args = parser.parse_args()

# --------------------- #

filename = args.f
if filename[-3:] != ".lt":
	filename = filename + ".lt"

type_bt = args.t
if type_bt != "rdm" and type_bt != "bal" and type_bt != "ill" :
	raise Exception("Not a valid type of tree") 

size = args.s

# --------------------- #

frdm = lambda : random.randint(1,101)

if type_bt == "rdm" :
	bt = generate_random_btree(frdm, size)

if type_bt == "bal" :
	bt = generate_balanced_btree(frdm, size)

if type_bt == "ill" :
	bt = generate_illbalanced_btree(frdm, size)

# --------------------- #

m = args.m
lt = LTree.init_from_bt(bt, m)
lt.write_file(filename)





from pyske.core.tree.ltree import LTree
from pyske.core.tree.ptree import PTree

from pyske.core.support.generate import *
from pyske.core.support.parallel import *

from pyske.examples.tree.prefix import prefix
from pyske.examples.tree.size import size
from pyske.examples.tree.size_by_node import size_by_node
from pyske.examples.tree.ancestors import ancestors

msg = "hello world!"
frdm = lambda : msg[random.randint(0,len(msg)-1)]

def test(bt):
	print (bt)
	at_root(lambda : print("-----"))

	m = 3
	lt = LTree.init_from_bt(bt, m)
	pt = PTree(lt)
	at_root(lambda : print(lt))
	print(pt)

	at_root(lambda : print("-----"))

	res = prefix(pt)
	at_root(lambda : print("prefix result:"))
	print(res)

	at_root(lambda : print("-----"))

	res = size(pt)
	at_root(lambda : print("size result:"))
	print(res)

	at_root(lambda : print("-----"))

	res = size_by_node(pt)
	at_root(lambda : print("size_by_node result:"))
	print(res)

	at_root(lambda : print("-----"))
	res = ancestors(pt)
	at_root(lambda : print("ancestors result:"))
	print(res)

msg = "hello world!"
frdm = lambda : msg[random.randint(0,len(msg)-1)]

if pid == 0:
	bal = generate_balanced_btree(frdm, 15)
	ill = generate_illbalanced_btree(frdm, 15)
	rdm = generate_random_btree(frdm, 15)
	for i in range(1, nprocs):
		comm.send({'b' : bal, 'i' : ill, 'r' : rdm}, dest=i, tag=1)
else:
	data = comm.recv(source=0, tag=1)
	bal = data['b']
	ill = data['i']
	rdm = data['r']

comm.Barrier()

at_root(lambda : print("\nBAlANCED\n"))
test(bal)

at_root(lambda :print("\nILL BAlANCED\n"))
test(ill)

at_root(lambda :print("\nRANDOM\n"))
test(rdm)





import pytest

from pyske.core.tree.ltree import parse_tag,TAG_LEAF,TAG_NODE,TAG_CRITICAL
from pyske.core.support.errors import UnknownTypeError

def test_parseVTag_leaf():
	tag = "L"
	exp = TAG_LEAF
	res = parse_tag(tag)
	assert exp == res


def test_parseVTag_node():
	tag = "N"
	exp = TAG_NODE
	res = parse_tag(tag)
	assert exp == res


def test_parseVTag_critical():
	tag = "C"
	exp = TAG_CRITICAL
	res = parse_tag(tag)
	assert exp == res


def test_parseVTag_unknown():
	tag = "_"
	with pytest.raises(UnknownTypeError):
		parse_tag(tag)



from pyske.core.list.slist import SList


def distribute_tree(lt, n):
	def sum3(x, y, z): return x+y+z
	total_size = lt.map(lambda x: 1, lambda x: 1).reduce(sum3, lambda x: x, sum3, sum3, sum3)
	avg_elements = int(total_size / n)

	# Definition of global_index and distribution 

	global_index = SList([])
	distribution = SList([])

	for i in range(n):
		distribution.append(0)

	current_pid = 0
	nb_segs = 1
	seg = lt[0]
	seg_length = seg.length()
	acc_size = seg_length
	global_index.append((0, seg_length))

	for seg_i in range(1, lt.length()):
		seg = lt[seg_i]
		curr_seg_length = seg.length()
		if current_pid == n - 1:
			# We need to give all the rest to the last processor
			if seg_i == lt.length() - 1:
				distribution[current_pid] = (nb_segs + 1)
				global_index.append((acc_size, curr_seg_length))
			else:
				nb_segs += 1
				curr_seg_length = seg.length()
				global_index.append((acc_size, curr_seg_length))
				acc_size = acc_size + curr_seg_length
		else:
			if seg_i == lt.length() - 1:
				distribution[current_pid] = (nb_segs + 1)
				global_index.append((acc_size, seg.length()))
			else:
				curr_seg_length = seg.length()
				if abs(avg_elements - (acc_size + curr_seg_length)) > abs(avg_elements - acc_size):
					distribution[current_pid] = nb_segs
					global_index.append((0, curr_seg_length))
					acc_size = curr_seg_length
					nb_segs = 1
					current_pid += 1
				else:
					nb_segs += 1
					global_index.append((acc_size, curr_seg_length))
					acc_size = acc_size + curr_seg_length

	return distribution, global_index



from pyske.core.list.slist import SList
from pyske.core.list.plist import PList
from pyske.core.support.parallel import *
from operator import *
from gc import *

iterations = 10


def test(f, input, name):
    times = SList()
    print(f'Test: {name}')
    for i in range(0, iterations):
        print(f'  Iteration: {i}', end='\r')
        collect()
        t = wtime()
        output = f(input)
        t = wtime() - t
        times.append(t)
    print(30 * ' ', end='\r')
    print(f'  Min time:\t{times.reduce(min)}')
    print(f'  Max time:\t{times.reduce(max)}')
    print(f'  Avg time:\t{times.reduce(add) / times.length()} ')
    return output


input = SList(range(1, 4_000_000))

t1 = test(lambda l: l.map2(l, add), input, "map2")

assert t1


from pyske.core.support.errors import NotEqualSizeError, EmptyError
from pyske.core.list.slist import SList
import pytest


# -------------------------- #

def test_head_empty():
    sl = SList()
    exp = None
    res = sl.head()
    assert res == exp


def test_head_one():
    sl = SList()
    exp = 1
    sl.append(exp)
    res = sl.head()
    assert res == exp


def test_head_several():
    sl = SList()
    exp = 1
    sl.append(exp)
    sl.append(exp + 1)
    res = sl.head()
    assert res == exp


# -------------------------- #

def test_tail_empty():
    sl = SList()
    exp = SList()
    res = sl.tail()
    assert res == exp


def test_tail_one():
    sl = SList([1])
    exp = SList()
    res = sl.tail()
    assert res == exp


def test_tail_several():
    sl = SList([1, 2, 3])
    exp = SList([2, 3])
    res = sl.tail()
    assert res == exp


# -------------------------- #

def test_length_nil():
    sl = SList()
    exp = 0
    res = sl.length()
    assert res == exp


def test_length_cons():
    sl = SList([1, 2, 3])
    exp = 3
    res = sl.length()
    assert res == exp


# -------------------------- #

def test_filter():
    sl = SList([1, 2, 3, 4, 5, 6, 7, 8])
    p = lambda x: x % 2 == 0
    res = sl.filter(p)
    for r in res:
        assert p(r)


# -------------------------- #

def test_empty_true():
    sl = SList()
    exp = True
    res = sl.empty()
    assert res == exp


def test_empty_false():
    sl = SList([1, 2, 3])
    exp = False
    res = sl.empty()
    assert res == exp


# -------------------------- #

def test_reverse_nil():
    sl = SList()
    exp = SList()
    res = sl.reverse()
    assert res == exp


def test_reverse_cons():
    sl = SList([1, 2, 3])
    exp = SList([3, 2, 1])
    res = sl.reverse()
    assert res == exp


# -------------------------- #

def test_map_empty():
    sl = SList()
    exp = SList()
    f = lambda x: x
    res = sl.map(f)
    assert res == exp


def test_map_inc():
    sl = SList([1, 2, 3])
    exp = SList([2, 3, 4])
    f = lambda x: x + 1
    res = sl.map(f)
    assert res == exp


def test_map_id():
    sl = SList([1, 2, 3])
    exp = SList([1, 2, 3])
    f = lambda x: x
    res = sl.map(f)
    assert res == exp


# -------------------------- #

def test_mapi_empty():
    sl = SList()
    exp = SList()
    f = lambda i, x: (i, x)
    res = sl.mapi(f)
    assert res == exp


def test_mapi_non_empty():
    sl = SList([1, 2, 3])
    exp = SList([0, 2, 6])
    f = lambda i, x: i * x
    res = sl.mapi(f)
    assert res == exp


def test_map_id():
    sl = SList([1, 2, 3])
    exp = SList([1, 2, 3])
    f = lambda i, x: x
    res = sl.mapi(f)
    assert res == exp

# -------------------------- #

def test_reduce_nil():
    e = 1232
    sl = SList()
    f = lambda x, y: x + y
    res = sl.reduce(f, e)
    exp = e
    assert res == exp


def test_reduce_cons():
    sl = SList([1, 2, 3, 4])
    f = lambda x, y: x + y
    res = sl.reduce(f)
    exp = 10
    assert res == exp


def test_reduce_sum_empty():
    sl = SList()
    f = lambda x, y: x + y
    exp = 0
    res = sl.reduce(f, 0)
    assert res == exp


def test_reduce_sum_non_empty():
    sl = SList([1, 2, 3, 4, 5, 6])
    f = lambda x, y: x + y
    exp = 22
    res = sl.reduce(f, 1)
    assert res == exp


# -------------------------- #

def test_scan_nil():
    c = 0
    sl = SList()
    f = lambda x, y: x + y
    res = sl.scan(f, c)
    exp = SList([0])
    assert res == exp


def test_scan_cons():
    c = 0
    sl = SList([1, 2, 3, 4])
    f = lambda x, y: x + y
    res = sl.scan(f, c)
    exp = SList([0, 1, 3, 6, 10])
    assert res == exp


# -------------------------- #

def test_scanr_empty():
    with pytest.raises(AssertionError):
        sl = SList()
        f = lambda x, y: x + y
        sl.scanr(f)


def test_scanr_singleton():
    sl = SList([1])
    f = lambda x, y: x + y
    res = sl.scanr(f)
    assert res == sl


def test_scanr_non_singleton():
    sl = SList([1, 2, 3, 4])
    f = lambda x, y: x + y
    res = sl.scanr(f)
    exp = SList([1, 3, 6, 10])
    assert res == exp


def test_scanr_full_distribution():
    sl = SList([(0, 1), (1, 3), (0, 3), (0, 3), (0, 3)])

    def f(x, y):
        (x1, y1) = x
        (x2, y2) = y
        return x1 + y1, y2

    res = sl.scanr(f)
    exp = SList([(0, 1), (1, 3), (4, 3), (7, 3), (10, 3)])
    assert res == exp


def test_scanr_full_distribution2():
        sl = SList([(0, 1), (1, 3), (4, 3)])

        def f(x, y):
            (x1, y1) = x
            (x2, y2) = y
            return x1 + y1, y2

        res = sl.scanr(f)
        exp = SList([(0, 1), (1, 3), (4, 3)])
        assert res == exp



# -------------------------- #

def test_scanl_empty():
    sl = SList([])
    f = lambda x, y: x + y
    res = sl.scanl(f, 0)
    exp = SList([])
    assert res == exp


def test_scanl_non_empty():
    sl = SList([1, 2, 3, 4])
    f = lambda x, y: x + y
    res = sl.scanl(f, 0)
    exp = SList([0, 1, 3, 6])
    assert res == exp

# -------------------------- #

def test_scanl_last_empty():
    sl = SList([])
    f = lambda x, y: x + y
    res = sl.scanl_last(f, 0)
    exp = ([], 0)
    assert res == exp


def test_scanl_last_non_empty():
    sl = SList([1, 2, 3, 4])
    f = lambda x, y: x + y
    res = sl.scanl_last(f, 0)
    exp = (SList([0, 1, 3, 6]), 10)
    assert res == exp


# -------------------------- #

def test_scanp_nil():
    c = 0
    sl = SList()
    f = lambda x, y: x + y
    res = sl.scanp(f, c)
    exp = SList([])
    assert res == exp

def test_scanp_cons():
    c = 0
    sl = SList([1, 2, 3, 4])
    f = lambda x, y: x + y
    res = sl.scanp(f, c)
    exp = SList([9, 7, 4, 0])
    assert res == exp


# -------------------------- #

def test_zip_nil():
    sl1 = SList()
    sl2 = SList()
    res = sl1.zip(sl2)
    exp = SList()
    assert res == exp


def test_zip_cons():
    sl1 = SList([1, 2, 3])
    sl2 = SList([2, 3, 4])
    res = sl1.zip(sl2)
    exp = SList([(1, 2), (2, 3), (3, 4)])
    assert res == exp


def test_zip_one_gt():
    sl1 = SList([1, 2, 3])
    sl2 = SList([2, 3])
    with pytest.raises(AssertionError):
        sl1.zip(sl2)


def test_zip_one_lt():
    sl1 = SList([2, 3])
    sl2 = SList([2, 3, 4])
    with pytest.raises(AssertionError):
        sl1.zip(sl2)

# -------------------------- #

def test_zipwith_nil():
    sl1 = SList()
    sl2 = SList()
    f = lambda x, y: x + y
    res = sl1.map2(f, sl2)
    exp = SList()
    assert res == exp


def test_zipwith_cons():
    sl1 = SList([1, 2, 3])
    sl2 = SList([2, 3, 4])
    f = lambda x, y: x + y
    res = sl1.map2(f, sl2)
    exp = SList([3, 5, 7])
    assert res == exp


def test_zipwith_one_gt():
    sl1 = SList([1, 2, 3])
    sl2 = SList([2, 3])
    f = lambda x, y: x + y
    with pytest.raises(AssertionError):
        sl1.map2(f, sl2)


def test_zipwith_one_lt():
    sl1 = SList([2, 3])
    sl2 = SList([2, 3, 4])
    f = lambda x, y: x + y
    with pytest.raises(AssertionError):
        sl1.map2(f, sl2)


# -------------------------- #

def test_from_str_simple():
    s = "[1;2;3]"
    res = SList.from_str(s)
    exp = SList([1, 2, 3])
    assert res == exp


def parser_tuple(s):
    s = s.replace("(", "")
    s = s.replace(")", "")
    ss = s.split(",")
    return (int(ss[0]), int(ss[1]))


def test_from_str_tuple():
    s = "[(1,2);(3,4)]"
    res = SList.from_str(s, parser=parser_tuple)
    exp = SList([(1, 2), (3, 4)])
    assert res == exp

# -------------------------- #



import pytest

from pyske.core.tree.btree import Leaf, Node
from pyske.core.support.errors import NotEqualSizeError

# -------------------------- #

def test_is_leaf_true():
	bt = Leaf(1)
	exp = True
	res = bt.is_leaf()
	assert exp == res


def test_is_leaf_false():
	bt = Node(1, Leaf(1), Leaf(1))
	exp = False
	res = bt.is_leaf()
	assert exp == res

# -------------------------- #

def test_is_node_false():
	bt = Leaf(1)
	exp = False
	res = bt.is_node()
	assert exp == res


def test_is_node_true():
	bt = Node(1, Leaf(1), Leaf(1))
	exp = True
	res = bt.is_node()
	assert exp == res

# -------------------------- #

def test_map_leaf():
	bt = Leaf(1)
	kl = lambda x : x + 1
	kn = lambda x : x - 1
	res = bt.map(kl, kn)
	exp = Leaf(2)
	assert exp == res


def test_map_node():
	bt = Node(1, Leaf(2), Leaf(3))
	kl = lambda x : x + 1
	kn = lambda x : x - 1
	res = bt.map(kl, kn)
	exp = Node(0, Leaf(3), Leaf(4))
	assert exp == res

# -------------------------- #

def test_mapt_leaf():
	bt = Leaf(1)
	kl = lambda x : x + 1
	kn = lambda x,y,z : max(x,max(y.get_value(),z.get_value()))
	res = bt.mapt(kl, kn)
	exp = Leaf(2)
	assert exp == res


def test_mapt_node():
	bt = Node(1, Leaf(2), Leaf(3))
	kl = lambda x : x + 1
	kn = lambda x,y,z : max(x,max(y.get_value(),z.get_value()))
	res = bt.mapt(kl, kn)
	exp = Node(3, Leaf(3), Leaf(4))
	assert exp == res

# -------------------------- #

def test_reduce_leaf():
	bt = Leaf(2)
	k = lambda x,y,z : max(x,max(y,z))
	res = bt.reduce(k)
	exp = 2
	assert exp == res


def test_reduce_node():
	bt = Node(1, Leaf(2), Leaf(3))
	k = lambda x,y,z : max(x,max(y,z))
	res = bt.reduce(k)
	exp = 3
	assert exp == res

# -------------------------- #

def test_uacc_leaf():
	bt = Leaf(1)
	k = lambda x,y,z : x + y + z
	res = bt.uacc(k)
	exp = Leaf(1)
	assert exp == res


def test_uacc_node():
	bt = Node(1, Leaf(2), Leaf(3))
	k = lambda x,y,z : x + y + z
	res = bt.uacc(k)
	exp =  Node(6, Leaf(2), Leaf(3))
	assert exp == res

# -------------------------- #

def test_dacc_leaf():
	c = 0
	bt = Leaf(1)
	gl = lambda x, y : x+y
	gr = lambda x, y : 0 if x-y < 0 else x-y
	res = bt.dacc(gl, gr, c)
	exp = Leaf(c)
	assert exp == res


def test_dacc_node():
	c = 0
	bt = Node(1, Node(2, Leaf(3), Leaf(4)), Leaf(5))
	gl = lambda x, y : x+y
	gr = lambda x, y : 0 if x-y < 0 else x-y
	res = bt.dacc(gl, gr, c)
	exp = Node(0, Node(1, Leaf(3), Leaf(0)), Leaf(0))
	assert exp == res

# -------------------------- #

def test_zip_leaf():
	bt1 = Leaf(1) 
	bt2 = Leaf(2)
	exp = Leaf((1,2))
	res = bt1.zip(bt2)
	assert exp == res


def test_zip_node():
	bt1 = Node(1, Leaf(2), Leaf(3))
	bt2 = Node(4, Leaf(5), Leaf(6))
	exp = Node((1,4), Leaf((2,5)), Leaf((3,6)))
	res = bt1.zip(bt2)
	assert exp == res


def test_zip_leaf_node():
	bt1 = Leaf(1) 
	bt2 = Node(4, Leaf(5), Leaf(6))
	with pytest.raises(AssertionError):
		bt1.zip(bt2)


def test_zip_node_leaf():
	bt1 = Node(1, Leaf(2), Leaf(3))
	bt2 = Leaf(2)
	with pytest.raises(AssertionError):
		bt1.zip(bt2)

# -------------------------- #

def test_zipwith_leaf():
	bt1 = Leaf(1) 
	bt2 = Leaf(2)
	f = lambda x,y : x + y
	exp = Leaf(3)
	res = bt1.map2(f, bt2)
	assert exp == res


def test_zipwith_node():
	bt1 = Node(1, Leaf(2), Leaf(3))
	bt2 = Node(4, Leaf(5), Leaf(6))
	f = lambda x,y : x + y
	exp = Node(5, Leaf(7), Leaf(9))
	res = bt1.map2(f, bt2)
	assert exp == res


def test_zipwith_leaf_node():
	bt1 = Leaf(1) 
	bt2 = Node(4, Leaf(5), Leaf(6))
	f = lambda x,y : x + y
	with pytest.raises(AssertionError):
		bt1.map2(f, bt2)


def test_zipwith_node_leaf():
	bt1 = Node(1, Leaf(2), Leaf(3))
	bt2 = Leaf(2)
	f = lambda x,y : x + y
	with pytest.raises(AssertionError):
		bt1.map2(f, bt2)

# -------------------------- #

def test_getchl_leaf():
	c = 1
	bt = Leaf(3)
	res = bt.getchl(c)
	exp = Leaf(c)
	assert res == exp


def test_getchl_node_right():
	c = 1
	bt = Node(3, Leaf(2), Node(4, Leaf(2), Leaf(6)))
	res = bt.getchl(c)
	exp = Node(2, Leaf(c), Node(2, Leaf(c), Leaf(c)))
	assert res == exp


def test_getchl_node_left():
	c = 1
	bt = Node(3, Node(4, Leaf(2), Leaf(6)), Leaf(2))
	res = bt.getchl(c)
	exp =  Node(4, Node(2, Leaf(c), Leaf(c)), Leaf(c))
	assert res == exp

# -------------------------- #

def test_getchr_leaf():
	c = 1
	bt = Leaf(3)
	res = bt.getchr(c)
	exp = Leaf(c)
	assert res == exp


def test_getchr_node_right():
	c = 1
	bt = Node(3, Leaf(2), Node(4, Leaf(2), Leaf(6)))
	res = bt.getchr(c)
	exp = Node(4, Leaf(c), Node(6, Leaf(c), Leaf(c)))
	assert res == exp


def test_getchr_node_left():
	c = 1
	bt = Node(3, Node(4, Leaf(2), Leaf(6)), Leaf(2))
	res = bt.getchr(c)
	exp =  Node(2, Node(6, Leaf(c), Leaf(c)), Leaf(c))
	assert res == exp


# -------------------------- #



from pyske.core.list.slist import SList
from pyske.core.list.plist import PList


def app(l1, l2):
    return SList(l1 + l2)


msg = "hello world!"


def test_init_to_seq_empty():
    pl = PList()
    res = pl.to_seq()
    exp = []
    assert res == exp


def test_init_to_seq_non_empty():
    pl = PList.init(lambda i: msg[i], len(msg))
    res = pl.to_seq()
    exp = list(msg)
    assert res == exp


def test_map_empty():
    f = lambda x: x.upper()
    input = PList()
    res = input.map(f).to_seq()
    exp = []
    assert res == exp


def test_map_non_empty():
    f = lambda x: x.upper()
    input = PList.init(lambda i: msg[i], len(msg))
    res = input.map(f).to_seq()
    exp = list(f(msg))
    assert res == exp


def test_mapi_empty():
    f = lambda i, x: f'{i}:{x.upper()}'
    input = PList()
    res = input.mapi(f).to_seq()
    exp = []
    assert res == exp


def test_mapi_non_empty():
    f = lambda i, x: f'{i}:{x.upper()}'
    input = PList.init(lambda i: msg[i], len(msg))
    res = input.mapi(f).to_seq()
    exp = SList(msg).mapi(f)
    assert res == exp


def test_scanr_non_empty():
    f = lambda x, y: x + y
    size = 23
    input = PList.init(lambda i: i, size)
    res = input.scanr(f).to_seq()
    exp = SList(range(0, size)).scanr(f)
    assert res == exp


def test_scanl_empty():
    f = lambda x, y: x + y
    size = 0
    input = PList.init(lambda i: i, size)
    res = input.scanl(f, 0).to_seq()
    exp = SList(range(0, size)).scanl(f, 0)
    assert res == exp


def test_scanl_non_empty():
    f = lambda x, y: x + y
    size = 23
    input = PList.init(lambda i: i, size)
    res = input.scanl(f, 0).to_seq()
    exp = SList(range(0, size)).scanl(f, 0)
    assert res == exp


def test_scanl_last_empty():
    f = lambda x, y: x + y
    size = 0
    input = PList.init(lambda i: i, size)
    res_pl, res_scalar = input.scanl_last(f, 0)
    res = (res_pl.to_seq(), res_scalar)
    exp = SList(range(0, size)).scanl_last(f, 0)
    assert res == exp


def test_scanl_last_non_empty():
    f = lambda x, y: x + y
    size = 23
    input = PList.init(lambda i: i, size)
    res_pl, res_scalar = input.scanl_last(f, 0)
    res = (res_pl.to_seq(), res_scalar)
    exp = SList(range(0, size)).scanl_last(f, 0)
    assert res == exp



from pyske.core.support.errors import ConstructorError
from pyske.core.list.slist import SList
from pyske.core.tree.btree import Node, Leaf, BTree


class RNode:
    """
    A class used to represent a Rose Tree (a tree with an arbitrary shape)

    ...

    Methods
    -------
    b2r(bt)
        Create a RNode instance from a BTree
    get_children()
        Get the children of the current RNode
    add_children(c)
        Add a children to the ones of the current RNode
    is_leaf()
        Indicates if the current RNode has no children, and then can be considered as a leaf
    is_node()
        Indicates if the current RNode has children, and then can be considered as a node
    get_value()
        Get the value of the current RNode
    set_value(v)
        Set a new value for the current RNode
    map(f)
        Applies a function to every values contained in the current instance
    reduce(f, g)
        Reduce the current instance into a single value using two operators
    uacc(f, g)
        Makes an upward accumulation of the values in a the current instance using two operators
    dacc(f, unit_f)
        Makes an downward accumulation of the values in a the current instance
    zip(rt)
        Zip the values contained in a second RTree with the ones in the current instance
    map2(rt, f)
        Zip the values contained in a second RTree with the ones in the current instance using a function
    racc(f, unit_f)
        Makes a rightward accumulation of the values in the current instance
    lacc(f, unit_f)
        Makes a leftward accumulation of the values in the current instance
    r2b()
        Get a BTree from the current instance
    """

    def __init__(self, value, ts=None):
        if isinstance(value, BTree):
            if value == Leaf(None):
                raise ConstructorError("A RTree cannot be constructed from a single Leaf that contains None")
            # Create a RTree from a BTree
            bt = value
            rt = RNode.b2r(bt)
            self.value = rt.get_value()
            self.children = rt.get_children()
        else:
            self.value = value
            if ts is None:
                self.children = SList([])
            else:
                self.children = SList(ts)

    @staticmethod
    def b2r(bt):
        """
        Create a RNode instance from a BTree
        """
        def aux(btree):
            if btree.is_leaf():
                val = btree.get_value()
                if val is None:
                    return SList()
                else:
                    return SList([RNode(val)])
            else:
                n = btree.get_value()
                left = btree.get_left()
                right = btree.get_right()
                res_l = aux(left)
                res_r = aux(right)
                res_head = RNode(n, res_l)
                res_r.insert(0, res_head)
                return res_r

        return aux(bt).head()

    def __str__(self):
        res = "rnode " + str(self.value) + "["
        ch = self.get_children()
        for i in range(0, self.get_children().length()):
            if i == ch.length() - 1:
                res = res + str(ch[i])
            else:
                res = res + str(ch[i]) + ", "
        return res + "]"

    def __eq__(self, other):
        if isinstance(other, RNode):
            ch1 = self.get_children()
            ch2 = other.get_children()
            if ch1.length() != ch2.length():
                return False
            for i in range(0, ch1.length()):
                if ch1[i] != ch2[i]:
                    return False
            return self.get_value() == other.get_value()
        return False

    def is_leaf(self):
        """Indicates if the current RNode has no children, and then can be considered as a leaf
        """
        return len(self.children) == 0

    def is_node(self):
        """Indicates if the current RNode has children, and then can be considered as a node
        """
        return len(self.children) != 0

    def get_children(self):
        """Get the children of the current RNode
        """
        return self.children

    def add_children(self, c):
        """Add a children to the ones of the current RNode

        Parameters
        ----------
        c
            The children to add
        """
        self.children.append(c)

    def get_value(self):
        """Get the value of the current RNode
        """
        return self.value

    def set_value(self, v):
        """Set a new value for the current RNode

        Parameters
        ----------
        v
            The new value to set
        """
        self.value = v

    def map(self, f):
        """Applies a function to every values contained in the current instance

        Parameters
        ----------
        f : callable
            The function to apply to every values of the current instance
        """
        v = f(self.get_value())
        # To each element of the list of children, we apply the RNode.map function
        ch = self.children.map(lambda x: x.map(f))
        return RNode(v, ch)

    def reduce(self, f, g):
        """Reduce the current instance into a single value using two operators

        Parameters
        ----------
        f : callable
            A binary operator to combine all sub reduction of the children of the current instance
            into an intermediate reduction
        g : callable
            A binary operator to combine the value of the current instance with the intermediate reduction
        """
        if self.children.empty():
            return self.get_value()
        # We calculate the reduction of each childen
        reductions = self.children.map(lambda x: x.reduce(f, g))
        # We combine every sub reductions using g
        red = reductions[0]
        for i in range(1, reductions.length()):
            red = g(red, reductions[i])
        # The final reduction is the result of the combination of sub reductions and the value of the current instance
        return f(self.get_value(), red)

    def uacc(self, f, g):
        """Makes an upward accumulation of the values in a the current instance using two operators

        Parameters
        ----------
        f : callable
            A binary operator to combine all top values from the accumulation within the children of the current
            instance into an intermediate accumulation
        g : callable
            A binary operator to combine the value of the current instance with the intermediate accumulation
        """
        v = self.reduce(f, g)
        ch = self.children.map(lambda x: x.uacc(f, g))
        return RNode(v, ch)

    def dacc(self, f, unit_f):
        """Makes an downward accumulation of the values in a the current instance

        Parameters
        ----------
        f : callable
            A function to accumulate the value of the current instance with the current accumulator
        unit_f
            A value such as, forall x, f(x, unit_f) = x
        """

        def dacc2(t, fct, c):
            # Auxiliary function to make an accumulation with an arbitrary accumulator
            return RNode(c, t.children.map(lambda x: dacc2(x, fct, fct(c, t.get_value()))))

        # Since the accumulator changes at each iteration, we need to use a changing parameter, not defined in dacc.
        # Use of an auxiliary function, with as a first accumulator, unit_f
        return dacc2(self, f, unit_f)

    def zip(self, rt):
        """Zip the values contained in a second RTree with the ones in the current instance

        Precondition
        -------------
        The lengths of self.children and rt.children should be equal

        Parameters
        ----------
        rt : :obj:`RTree`
            The RTree to zip with the current instance
        """
        ch1 = self.get_children()
        ch2 = rt.get_children()
        assert ch1.length() == ch2.length(), "The rose trees cannot be zipped (not the same shape)"
        ch = SList([])
        for i in range(0, ch1.length()):
            ch.append(ch1[i].zip(ch2))
        v = (self.get_value(), rt.get_value())
        return RNode(v, ch)

    def map2(self, rt, f):
        """Zip the values contained in a second RTree with the ones in the current instance using a function

        Precondition
        -------------
        The lengths of self.children and rt.children should be equal

        Parameters
        ----------
        rt : :obj:`RTree`
            The RTree to zip with the current instance
        f : callable
            A function to zip values
        """
        ch1 = self.get_children()
        ch2 = rt.get_children()
        assert ch1.length() == ch2.length(), "The rose trees cannot be zipped (not the same shape)"
        ch = SList([])
        for i in range(0, ch1.length()):
            ch.append(ch1[i].map2(ch2, f))
        v = f(self.get_value(), rt.get_value())
        return RNode(v, ch)

    def racc(self, f, unit_f):
        """Makes a rightward accumulation of the values in the current instance

        rAcc (+) (RNode a ts)
            = let rs = scanl (+) [root ts[i] | i in [1 .. #ts]]
            in  RNode unit_(+) [setroot (rAcc (+) ts[i]) r[i] | i in [1 .. #ts]]

        Parameters
        ----------
        f : callable
            A function to accumulate the value of the current instance with the current accumulator
        unit_f
            A value such as, forall x, f(x, unit_f) = x
        """

        rv = self.get_children().map(lambda x: x.get_value())
        rs = rv.scanl(f, unit_f)
        ch = SList()
        ch0 = self.get_children()
        for i in range(0, ch0.length()):
            c = ch0[i]
            cs = c.racc(f, unit_f)
            cs.set_value(rs[i])
            ch.append(cs)
        return RNode(unit_f, ch)

    def lacc(self, f, unit_f):
        """Makes a leftward accumulation of the values in the current instance

        lAcc (+) (RNode a ts)
            = let rs = scanp (+) [root ts[i] | i in [1 .. #ts]]
            in  RNode unit_(+) [setroot (lAcc (+) ts[i]) r[i] | i in [1 .. #ts]]

        Parameters
        ----------
        f : callable
            A function to accumulate the value of the current instance with the current accumulator
        unit_f
            A value such as, forall x, f(x, unit_f) = x
        """
        rv = self.get_children().map(lambda x: x.get_value())
        rs = rv.scanp(f, unit_f)
        ch = SList()
        ch0 = self.get_children()
        for i in range(0, ch0.length()):
            c = ch0[i]
            cs = c.lacc(f, unit_f)
            cs.set_value(rs[i])
            ch.append(cs)
        return RNode(unit_f, ch)

    def r2b(self):
        """Get a BTree from the current instance
        """
        def r2b1(t, ss):
            a = t.get_value()
            left = r2b2(t.get_children())
            right = r2b2(ss)
            return Node(a, left, right)

        def r2b2(ts):
            if ts.empty():
                return Leaf(None)
            else:
                h = ts.head()
                t = ts.tail()
                return r2b1(h, t)

        return r2b1(self, SList())



import pytest

from pyske.core.tree.rtree import RNode
from pyske.core.tree.btree import Node, Leaf 
from pyske.core.list.slist import SList
from pyske.core.support.errors import ConstructorError

# -------------------------- #

def test_b2r_leaf_none():
	bt = Leaf(None)
	with pytest.raises(ConstructorError):
		RNode(bt)


def test_b2r_leaf():
	bt = Leaf(1)
	res = RNode(bt)
	exp = RNode(1)
	assert res == exp


def test_b2r_node_from_rt():
	bt = Node(1, 
		Node(2, 
			Leaf(None), 
			Node(3, 
				Node(5, 
					Leaf(None), 
					Node(6, 
						Leaf(None), 
						Leaf(None)
						)
					), 
				Node(4, 
					Leaf(None), 
					Leaf(None)
					)
				)
			), 
		Leaf(None)
		)
	rn5 = RNode(5)
	rn6 = RNode(6)
	rn3 = RNode(3, SList([rn5, rn6]))
	rn2 = RNode(2)
	rn4 = RNode(4)
	exp = RNode(1, SList([rn2, rn3, rn4]))
	res = RNode(bt)

	assert res == exp

# -------------------------- #

def test_r2b_1():
	ch = SList()
	ch.append(RNode(2))
	ch.append(RNode(3))
	rn = RNode(1, ch)
	res = rn.r2b()
	exp = Node(1, Node(2, Leaf(None), Node(3,Leaf(None), Leaf(None))), Leaf(None))
	assert res == exp


def test_r2b_2():
	rn5 = RNode(5)
	rn6 = RNode(6)
	rn3 = RNode(3, SList([rn5, rn6]))
	rn2 = RNode(2)
	rn4 = RNode(4)
	rn1 = RNode(1, SList([rn2, rn3, rn4]))
	res = rn1.r2b()
	exp = \
	Node(1, 
		Node(2, 
			Leaf(None), 
			Node(3, 
				Node(5, 
					Leaf(None), 
					Node(6, 
						Leaf(None), 
						Leaf(None)
						)
					), 
				Node(4, 
					Leaf(None), 
					Leaf(None)
					)
				)
			), 
		Leaf(None)
		)
	assert res == exp


def test_map():
    rt = RNode(15, [RNode(24),
                    RNode(32, [RNode(56),
                               RNode(63)]),
                    RNode(41)])
    res = rt.map(lambda x: str(x))
    exp = RNode("15", [RNode("24"),
                       RNode("32", [RNode("56"),
                                    RNode("63")]),
                       RNode("41")])


def test_reduce():
    rt = RNode(15, [RNode(24),
                    RNode(32, [RNode(56),
                               RNode(63)]),
                    RNode(41)])
    plus2 = lambda x, y: x+y
    times2 = lambda x, y: x*y
    res = rt.reduce(plus2,times2)
    exp =  3503055
    assert res == exp


def test_uacc():
    rt = RNode(15, [RNode(24),
                    RNode(32, [RNode(56),
                               RNode(63)]),
                    RNode(41)])
    plus2 = lambda x, y: x+y
    times2 = lambda x, y: x*y
    res = rt.uacc(plus2,times2)
    exp =  RNode(3503055, [RNode(24),
                           RNode(3560,[RNode(56),
                                     RNode(63)]),
                           RNode(41)])
    assert res == exp


def test_dacc():
    rt = RNode(15, [RNode(24),
                    RNode(32, [RNode(56),
                               RNode(63)]),
                    RNode(41)])
    plus2 = lambda x, y: x+y
    res = rt.dacc(plus2,0)
    exp =  RNode(0, [RNode(15),
                    RNode(15,[RNode(47),
                              RNode(47)]),
                    RNode(15)])
    assert res == exp


def test_lacc():
    rt = RNode(15, [RNode(24),
                    RNode(32, [RNode(56),
                               RNode(63)]),
                    RNode(41)])
    plus2 = lambda x, y: x+y
    res = rt.lacc(plus2, 0)
    exp =  RNode(0, [RNode(73),
                    RNode(41,[RNode(63),
                              RNode(0)]),
                    RNode(0)])
    assert res == exp


def test_racc():
    rt = RNode(15, [RNode(24),
                    RNode(32,[RNode(56),
                              RNode(63)]),
                    RNode(41)])
    plus2 = lambda x, y: x+y
    res = rt.racc(plus2, 0)
    exp =  RNode(0, [RNode(0),
                    RNode(24,
                        [RNode(0),
                         RNode(56)]),
                    RNode(56)])
    assert res == exp



from pyske.core.tree.btree import BTree, Leaf, Node
from pyske.core.list.slist import SList

import random


def generate_random_list(frdm, size):
	res = SList([])
	for i in range(size):
		res.append(frdm())
	return res


def generate_balanced_btree(frdm, size):
	current_size = 1
	bt = Leaf(frdm())
	while current_size < size:
		bt2 = bt.map(lambda x: frdm(), lambda x: frdm())
		bt = Node(frdm(), bt, bt2)
		current_size = 2 * current_size + 1
	return bt


def __insert_values_btree(v1, v2, bt):
	if bt.is_leaf():
		return Node(bt.get_value(), Leaf(v1), Leaf(v2))
	else:  # bt.is_node()
		rdm = random.randint(1, 101)
		if rdm > 50:
			# insert left
			left = __insert_values_btree(v1, v2, bt.get_left())
			right = bt.get_right()
			return Node(bt.get_value(), left, right)
		else:
			# insert right
			left = bt.get_left()
			right = __insert_values_btree(v1, v2, bt.get_right())
			return Node(bt.get_value(), left, right)
			

def generate_random_btree(frdm, size):
	adj_size = (size if size % 2 == 1 else size + 1)
	values = generate_random_list(frdm, adj_size)
	bt = Leaf(values[0])
	for i in range(1, len(values), 2):
		bt = __insert_values_btree(values[i], values[i+1], bt)
	return bt


def generate_illbalanced_btree(frdm, size):
	current_size = 1
	bt = Leaf(frdm())
	while current_size < size:
		bt = Node(frdm(), Leaf(frdm()), bt)
		current_size = current_size + 2
	return bt



from mpi4py import MPI


comm = MPI.COMM_WORLD
pid = comm.Get_rank()
nprocs = comm.Get_size()


def wtime():
    return MPI.Wtime()


def local_size_pid(idp, size):
    return int(size / nprocs) + (1 if idp < size % nprocs else 0)


def local_size(size):
    return local_size_pid(pid, size)


def at_root(f):
    if pid == 0:
        return f()
    else:
        return None


def scan(op, x):
    xs = comm.alltoall([x for _ in range(0, nprocs)])
    pre = xs[0]
    for i in range(1, pid):
        pre = op(pre, xs[i])
    res = xs[pid]
    for i in range(pid + 1, nprocs):
        res = op(res, xs[i])
    if pid != 0:
        return pre, op(pre, res)
    else:
        return pre, res


def get_conc_reader(filename):
    return MPI.File.Open(comm, filename, MPI.MODE_RDONLY)


def close_conc_file(file):
    return MPI.Close(file)


def run_ptests(tests, filename=""):
    for f in tests:
        try:
            f()
            at_root(lambda: print("\033[32m[OK] " + str(f)[10:][:-16] + " (test/" + str(filename) + ".py)\033[0m"))
        except Exception as e:
            at_root(lambda: print("\033[31m[KO] " + str(f) + " (test/" + str(filename) + ".py)\033[0m"))
            at_root(lambda: print(e))



import sys

from pyske.core.support.errors import EmptyError, UnknownTypeError, IllFormedError, ApplicationError, NotSameTagError
from pyske.core.list.slist import SList
from pyske.core.tree.btree import Leaf, Node


MINUS_INFINITY = int((-sys.maxsize - 1) / 2)
TAG_LEAF = 1
TAG_NODE = 2
TAG_CRITICAL = 3

SEPARATOR_TV = "^"
LEFT_TV = "("
RIGHT_TV = ")"
LEFT_SEG = "["
RIGHT_SEG = "]"
SEPARATOR_SEG = ";"
EXT_FILE_LT = "lt"


def parse_tag(vt):
    """
    Get a VTag from a label

    Parameters
    ----------
    vt : str
        A label that we want to get the VTag

    Raises
    ------
    UnknownTypeError
        If the label is not known in VTag
    """
    vt = vt.strip()
    if vt == "C":
        return TAG_CRITICAL
    if vt == "L":
        return TAG_LEAF
    if vt == "N":
        return TAG_NODE
    raise UnknownTypeError('Type of value unknown: ' + str(vt))


class TaggedValue:
    """
    A class used to represent a Value in a linearized tree

    ...

    Attributes
    ----------
    val
        A value describing the current element
    tag : int
        A integer value to describe the current value

    Methods
    -------
    get_tag()
        Get the tag value used to describe the tag of the current instance
    getValue()
        Get the value contained in the current instance
    is_leaf()
        Indicates if the current instance is tagged by the Leaf tag
    is_critical()
        Indicates if the current instance is tagged by the Critical tag
    is_node()
        Indicates if the current instance is tagged by the Node tag
    """

    def __init__(self, val, t):
        self.val = val
        if (t == TAG_LEAF) | (t == TAG_NODE) | (t == TAG_CRITICAL):
            self.tag = t
        else:
            try:
                self.tag = parse_tag(t)
            except UnknownTypeError as e:
                print(str(e))

    def __str__(self):
        def tag2str(tag):
            if tag == TAG_CRITICAL:
                return "C"
            if tag == TAG_LEAF:
                return "L"
            if tag == TAG_NODE:
                return "N"

        return LEFT_TV + str(self.val) + SEPARATOR_TV + tag2str(self.tag) + RIGHT_TV

    def __eq__(self, other):
        if isinstance(other, TaggedValue):
            return (self.tag == other.tag) and (self.val == other.val)
        return False

    def get_tag(self):
        """Get the tag value used to describe the tag of the current instance
        """
        return self.tag

    def get_value(self):
        """Get the tag of the current instance
        """
        return self.val

    def is_leaf(self):
        """Indicates if the current instance is tagged by the Leaf tag
        """
        return self.tag == TAG_LEAF

    def is_critical(self):
        """Indicates if the current instance is tagged by the Critical tag
        """
        return self.tag == TAG_CRITICAL

    def is_node(self):
        """Indicates if the current instance is tagged by the Node tag
        """
        return self.tag == TAG_NODE


class Segment(SList):
    """A list of TaggedValue

    ...


    Methods
    -------
    has_critical()
        Indicates if the current instance contains a value tagged by the Critical VTag
    map_local(kl, kn)
        Applies function kl to each leaf and function kn to each internal node
        and the m-critical node in a local Segment
    reduce_local(k, phi, psi_l, psi_r)
        Reduces a local Segment into a value
    reduce_global(psi_n)
        Makes a global reduction using local reductions of Segments
    uacc_local(k, phi, psi_l, psi_r)
        Computes local upwards accumulation and reduction.
    uacc_global(psi_n)
        Performs sequential upwards accumulation
    uacc_update(seg, k, lc, rc)
        Makes an update of the current accumulation, using initial values and the top accumulated values
    dacc_path(phi_l, phi_r, psi_u)
        Finds the m-critical node and then computes two values only on the path from the root node
        to the m-critical node
    dacc_global(psi_d, c)
        Performs sequential downwards accumulation
    dacc_local(gl, gr, c)
        Computes local downward accumulation for the current instance using an accumulative parameter
        resulting of a global downward accumulation
    from_str(s, parser)
        Get a segment from a string value
    zip()
        Zip the values contained in a second Segment with the ones in the current instance
    map2(f, seg)
        Zip the values contained in a second Segment with the ones in the current instance using a function
    """

    def __eq__(self, other):
        if isinstance(other, Segment):
            if self.length() != other.length():
                return False
            for i in range(0, self.length()):
                if self[i] != other[i]:
                    return False
            return True
        return False

    def has_critical(self):
        """Indicates if the current instance contains a value tagged by the Critical tag
        """
        for v in self:
            if v.is_critical():
                return True
        return False

    def map_local(self, kl, kn):
        """Applies function kl to each leaf and function kn to each internal node
        and the m-critical node in a local Segment


        Parameters
        ----------
        kl : callable
            The function to apply to every values tagged by LEAF of the current instance
        kn : callable
            The function to apply to every values tagged by CRITICAL or NODE of the current instance
        """
        res = Segment([None] * self.length())
        for i in range(self.length()):
            tv = self[i]
            if tv.is_leaf():
                v = TaggedValue(kl(tv.get_value()), tv.get_tag())
            else:  # v.is_node() or v.is_critical()
                v = TaggedValue(kn(tv.get_value()), tv.get_tag())
            res[i] = v
        return res

    def reduce_local(self, k, phi, psi_l, psi_r):
        """Reduces a local Segment into a value

        Precondition
        -------------
        self should not be empty

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        phi : callable
            A function used to respect the closure property
        psi_l : callable
            A function used to respect the closure property to make partial computation on the left
        psi_r : callable
            A function used to respect the closure property to make partial computation on the right

        Raises
        ------
        IllFormedError
            If the current instance does not represent a correct linearized subtree
            That is there is a node that does not have two children which can be either a leaf value or a critical value
        """
        assert not self.empty(), "reduce_local cannot be applied to an empty Segment"
        stack = []
        d = MINUS_INFINITY
        has_critical = False
        for v in self.reverse():
            # Starts by the end, that is the most deep leaves
            # We stack every elements we already reduced
            if v.is_leaf():
                # We cannot reduce a leaf value
                stack.append(v.get_value())
                d = d + 1
            elif v.is_node():
                if len(stack) < 2:
                    raise IllFormedError(
                        "reduce_local cannot be applied if there is a node that does not have"
                        "two children in the current instance")
                # We get two sub-reductions to make a reduction with the current node value
                lv = stack.pop()
                rv = stack.pop()
                if d == 0:
                    # The current node is an ancestor of a critical value by the left
                    # That is, there is a critical value on its left children in a BTree representation
                    # We process and stack a partial reduction
                    stack.append(psi_l(lv, phi(v.get_value()), rv))
                elif d == 1:
                    # The current node is an ancestor of a critical value by the right
                    # That is, there is a critical value on its right children in a BTree representation
                    # We process and stack a partial reduction
                    stack.append(psi_r(lv, phi(v.get_value()), rv))
                    d = 0
                else:
                    # We did not meet a critical value, we process and stack a normal reduction
                    stack.append(k(lv, v.get_value(), rv))
            else:  # v.is_critical()
                # we process and stack the reduction of critical value
                stack.append(phi(v.get_value()))
                has_critical = True
                d = 0
        top = stack.pop()
        if has_critical:
            # The current instance represented a node in the global structure of a linearized tree
            return TaggedValue(top, "N")
        else:
            # The current instance represented a leaf in the global structure of a linearized tree
            return TaggedValue(top, "L")

    def reduce_global(self, psi_n):
        """Makes a global reduction using local reductions of Segments

        Precondition
        -------------
        self should empty, and should not have critical nodes

        Parameters
        ----------
        psi_n : callable
            A function used to respect the closure property on k (the initial function used for reduction)
            to allow partial computation

        Raises
        ------
        IllFormedError
            If the current instance does not represent a list of top values of a correct list of subtrees
            That is for each node, there is not exist two children, of there is a critical value in the current instance
        """
        assert not self.has_critical(), "reduce_global cannot be applied to a Segments which contains a critical"
        assert not self.empty(), "reduce_global cannot be applied to an empty Segment"
        stack = []
        for g in self.reverse():
            # We stack every value we already reduced
            if g.is_leaf():
                # Nothing to calculate, we only stack the value
                stack.append(g.get_value())
            else:  # g.is_node()
                # We get two sub reductions to make a total reduction of the current node
                if len(stack) < 2:
                    raise IllFormedError(
                        "reduce_global cannot be applied if there is a node that does not have two children "
                        "in the current instance")
                lv = stack.pop()
                rv = stack.pop()
                # We process and stack a reduction
                stack.append(psi_n(lv, g.get_value(), rv))
        top = stack.pop()
        return top

    def uacc_local(self, k, phi, psi_l, psi_r):
        """Computes local upwards accumulation and reduction

        Precondition
        -------------
        self should empty

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        phi : callable
            A function used to respect the closure property
        psi_l : callable
            A function used to respect the closure property to make partial computation on the left
        psi_r : callable
            A function used to respect the closure property to make partial computation on the right

        Raises
        ------
        IllFormedError
            If the current instance does not represent a correct linearized subtree
            That is there is a node that doesn't have two children which can be either a leaf value or a critical value
        """
        assert not self.empty(), "uacc_local cannot be applied to an empty Segment"
        stack = []
        d = MINUS_INFINITY
        res = Segment([None] * self.length())
        has_crit = False
        for i in reversed(range(self.length())):
            v = self[i]
            # We stack all the values of previous accumulation
            if v.is_leaf():
                res[i] = v
                stack.append(v.get_value())
                d = d + 1

            elif v.is_node():
                if len(stack) < 2:
                    raise IllFormedError(
                        "uacc_local cannot be applied if there is a node that does not have two children "
                        "in the current instance")
                # We get the values of two sub upward accumulation
                lv = stack.pop()
                rv = stack.pop()
                if d == 0:
                    # The current node is an ancestor of a critical value by the left
                    # That is, there is a critical value on its left children in a BTree representation
                    # We process and stack the value of a partial accumulation
                    val = phi(v.get_value())
                    stack.append(psi_l(lv, val, rv))
                    res[i] = None
                elif d == 1:
                    # The current node is an ancestor of a critical value by the left
                    # That is, there is a critical value on its left children in a BTree representation
                    # We process and stack the value of a partial accumulation
                    val = phi(v.get_value())
                    stack.append(psi_r(lv, val, rv))
                    res[i] = None
                    d = 0
                else:
                    # We did not meet a critical value, we can process a normal upward accumulation with k
                    val = k(lv, v.get_value(), rv)
                    res[i] = TaggedValue(val, v.get_tag())
                    stack.append(val)
                    d = d - 1

            else:  # v.is_critical()
                # The current value is critical. We make a partial accumulation with phi and stack the result
                stack.append(phi(v.get_value()))
                res[i] = None
                d = 0
                has_crit = True

        top = stack.pop()
        tag = "N" if has_crit else "L"
        # We return both the top values for following global upward accumulation, and the current accumulated subtree
        return TaggedValue(top, tag), res

    def uacc_global(self, psi_n):
        """Performs sequential upwards accumulation

        Precondition
        -------------
        self should not have critical nodes

        Parameters
        ----------
        psi_n : callable
            A function used to respect the closure property on k (the initial function used for accumulation)
            to allow partial computation

        Raises
        ------
        IllformedError
            If the current instance does not represent a correct linearized subtree
            That is there is a node that doesn't have two children
        """
        assert not self.has_critical(), "uacc_global cannot be applied to a Segments which contains a critical"

        stack = []
        res = Segment([None] * self.length())
        for i in reversed(range(self.length())):
            g = self[i]
            # We process a global accumulation using a stack to store previous accumulation,
            # to get them for the accumulation on nodes
            if g.is_leaf():
                res[i] = g
                val = g.get_value()
            else:  # g.is_node()
                if len(stack) < 2:
                    raise IllFormedError(
                        "uacc_global cannot be applied if there is a node that does not have two children "
                        "in the current instance")
                lv = stack.pop()
                rv = stack.pop()
                val = psi_n(lv, g.get_value(), rv)
                res[i] = TaggedValue(val, g.get_tag())
            stack.append(val)
        # We get the top value of the accumulation
        return res

    def uacc_update(self, seg2, k, lc, rc):
        """Makes an update of the current accumulation, using initial values and the top accumulated values

        Precondition
        -------------
        The lengths of self and seg2 should be equal

        Parameters
        ----------
        seg2 : :obj:`Segment`
            Result of a local accumulation
        k : callable
            The function used to reduce a BTree into a single value
        lc
            Top value of the left children in a global structure
        rc
            Top value of the left children in a global structure

        Raises
        ------
        IllFormedError
            If the current instance does not represent a correct linearized subtree
            That is there is a node that doesn't have two children
        """
        assert self.length() == seg2.length(), "uacc_update cannot needs to Segment of same size as input"

        stack = [rc, lc]
        d = MINUS_INFINITY
        res = Segment([None] * self.length())
        for i in reversed(range(seg2.length())):
            v1 = self[i]
            v2 = seg2[i]
            # We update the accumulation from seg2

            # We stack the values already updated to process updates on nodes

            if v1.is_leaf():
                # The result of the accumulation is the node made in seg2
                res[i] = v2
                stack.append(v2.get_value())
                d = d + 1

            elif v1.is_node():
                if len(stack) < 2:
                    raise IllFormedError(
                        "uacc_update cannot be applied if there is a node that does not have two children "
                        "in the current instance")
                # We need two sub accumulation values to process the accumulation of a node
                lv = stack.pop()
                rv = stack.pop()
                if d == 0 or d == 1:
                    # We met a critical value before, so the accumulation is not completed yet
                    val = k(lv, v1.get_value(), rv)
                    res[i] = TaggedValue(val, v1.get_tag())
                    stack.append(val)
                    d = 0
                else:
                    # We did not meet a critical value before, so the accumulation is completed yet
                    res[i] = v2
                    stack.append(v2.get_value())
                    d = d - 1
            else:  # v1.is_critical()
                if len(stack) < 2:
                    raise IllFormedError(
                        "uacc_update cannot be applied if there is a node that does not have two children "
                        "in the current instance")
                # We need two sub accumulation values to process the accumulation of a critical node
                lv = stack.pop()
                rv = stack.pop()
                val = k(lv, v1.get_value(), rv)
                res[i] = TaggedValue(val, v1.get_tag())
                stack.append(val)
                d = 0
        return res

    def dacc_path(self, phi_l, phi_r, psi_u):
        """Finds the critical node and then computes two values only on the path from the root node to the critical node

        Precondition
        -------------
        self should not be empty

        Parameters
        ----------
        phi_l : callable
            A function used to respect the closure property to allow partial computation on the left
        phi_r : callable
            A function used to respect the closure property to allow partial computation on the right
        psi_u : callable
            A function used to respect the closure property to make partial computation

        Raises
        ------
        ApplicationError
            If the current instance does not contain a critical value
        """
        assert not self.empty(), "dacc_path cannot be applied to an empty Segment"
        d = MINUS_INFINITY
        # The value to pass to the left children for a total downward accumulation
        to_l = None
        # The value to pass to the right children for a total downward accumulation
        to_r = None
        has_critical = False
        for v in self.reverse():
            if v.is_leaf():
                d = d + 1
            elif v.is_node():
                if d == 0:
                    # The current node is an ancestor of a critical value by the left
                    # That is, there is a critical value on its left children in a BTree representation
                    # We process and stack the value of a partial downward accumulation
                    to_l = psi_u(phi_l(v.get_value()), to_l)
                    to_r = psi_u(phi_l(v.get_value()), to_r)
                elif d == 1:
                    # The current node is an ancestor of a critical value by the right
                    # That is, there is a critical value on its right children in a BTree representation
                    # We process and stack the value of a partial downward accumulation
                    to_l = psi_u(phi_l(v.get_value()), to_l)
                    to_r = psi_u(phi_l(v.get_value()), to_r)
                    d = 0
                else:
                    d = d - 1
            else:  # v.is_critical()
                has_critical = True
                to_l = phi_l(v.get_value())
                to_r = phi_r(v.get_value())
                d = 0
        if not has_critical:
            raise ApplicationError("dacc_path must be imperatively applied to a Segment which contains a critical node")
        return TaggedValue((to_l, to_r), "N")

    def dacc_global(self, psi_d, c):
        """Performs sequential downwards accumulation

        Precondition
        -------------
        self should bot have critical nodes

        Parameters
        ----------
        psi_d : callable
            A function used to respect the closure property on gr and gl (initial functions used for up accumulation)
            to make partial downward accumulation
        c
            Initial value of the accumulator

        Raises
        ------
        IllFormedError
            If the current instance does not represent a correct linearized subtree
            That is there are several leaves that doesn't have a parent in a BTree representation
        """
        stack = [c]
        res = Segment([None] * self.length())
        assert not self.has_critical(), "dacc_global cannot be applied to Segment which contains a critical node"

        for i in range(self.length()):
            v = self[i]
            if len(stack) == 0:
                raise IllFormedError(
                    "dacc_global cannot be applied to ill-formed Segments that is two leaf values do not have a parent")
            # We add the previous accumulation as a new value of our result
            val = stack.pop()
            res[i] = TaggedValue(val, v.get_tag())

            # If the current value is node, we need to update the value to pass to the right, and left children
            # These values are contained in the stack
            if v.is_node():
                (to_l, to_r) = v.get_value()
                stack.append(psi_d(val, to_r))
                stack.append(psi_d(val, to_l))
        return res

    def dacc_local(self, gl, gr, c):
        """Computes local downward accumulation for the current instance using an accumulative parameter resulting of a
        global downward accumulation

        Parameters
        ----------
        gl : callable
            Function to make a downward accumulation to the left
        gr : callable
            Function to make a downward accumulation to the right
        c
            Initial value of the accumulator

        Raises
        ------
        IllFormedError:
            If the current instance does not represent a correct linearized subtree
            That is there are several leaves that doesn't have a parent in a BTree representation or if there is not a
            value to accumulate from above
        """
        # We update not finished accumulation locally using the value from the parent in the global representation of
        # a linearized tree
        stack = [c]
        res = Segment([None] * self.length())
        for i in range(self.length()):
            v = self[i]
            if v.is_leaf() or v.is_critical():
                if len(stack) == 0:
                    raise IllFormedError(
                        "dacc_local cannot be applied if there are two leaf values, or critical values that"
                        "do not have a parent")
                # We get the accumulated value passed from the last parent
                val = stack.pop()
                res[i] = TaggedValue(val, v.get_tag())
            else:  # v.is_node()
                if len(stack) == 0:
                    raise IllFormedError(
                        "dacc_local cannot be applied if there is not a value to accumulate from above")
                val = stack.pop()
                # We get the accumulated value passed from the last parent
                # And two new ones, one for the left children, and one to the right, using the gr and gl functions
                res[i] = TaggedValue(val, v.get_tag())
                stack.append(gr(val, v.get_value()))
                stack.append(gl(val, v.get_value()))
        return res

    def get_left(self, i):
        """Get the left children of a value at the i-th index

        Precondition
        -------------
        self should not have critical nodes
        The value at the looked index should not be a leaf, and should not be the last element

        Parameters
        ----------
        i : int
            The index of the value that we want to get the left children
        """
        assert not self.has_critical(), "The left children of a value in a non-global Segment cannot be found"
        assert not self[i].is_leaf(), "A leaf value doesn't have a left children"
        assert i < self.length() - 1, "Cannot get the left children of a node in an ill-formed Segment"
        return self[i + 1]

    def get_right(self, i):
        """Get the right children of a value at the i-th index

        Precondition
        -------------
        self should not have critical nodes
        The value at the looked index should not be a leaf, and should not be the before last element

        Parameters
        ----------
        i : int
            The index of the value that we want to get the right children
        """
        assert not self.has_critical(), "The right children of a value in a non-global Segment cannot be found"
        assert not self[i].is_leaf(), "A leaf value doesn't have a right children"
        assert i < self.length() - 2, "Cannot get the left children of a node in an ill-formed Segment"

        def get_right_index(gt, idx):
            if gt[idx + 1].is_leaf():
                return idx + 2
            else:
                return 1 + get_right_index(gt, idx + 1)

        j = get_right_index(self, i)
        return self[j]

    def zip(self, seg):
        """Zip the values contained in a second Segment with the ones in the current instance

        Precondition
        -------------
        The lengths of self and seg should be equal

        Parameters
        ----------
        seg : :obj:`Segment`
            The Segment to zip with the current instance

        Raises
        ------
        NotSameTagError
            If two values with not the same tag are trying to be zipped together
        """
        assert self.length() == seg.length(), "The linearized trees have not the same shape"
        res = Segment([None] * self.length())
        for j in range(self.length()):
            tv1 = self[j]
            tv2 = seg[j]
            if tv1.get_tag() != tv2.get_tag():
                raise NotSameTagError("Two zipped values have not the same tag")
            tv = TaggedValue((tv1.get_value(), tv2.get_value()), tv1.get_tag())
            res[j] = tv
        return res

    def map2(self, f, seg):
        """Zip the values contained in a second Segment with the ones in the current instance using a function

        Precondition
        -------------
        The lengths of self and seg should be equal

        Parameters
        ----------
        f : callable
            A function to zip values
        seg : :obj:`Segment`
            The Segment to zip with the current instance

        Raises
        ------
        NotSameTagError
            If two values with not the same tag are trying to be zipped together
        """
        assert self.length() == seg.length(), "The linearized trees have not the same shape"
        res = Segment([None] * self.length())
        for j in range(self.length()):
            tv1 = self[j]
            tv2 = seg[j]
            if tv1.get_tag() != tv2.get_tag():
                raise NotSameTagError("Two zipped values have not the same tag")
            tv = TaggedValue(f(tv1.get_value(), tv2.get_value()), tv1.get_tag())
            res[j] = tv
        return res

    @staticmethod
    def from_str(s, parser=int):
        """Get a segment from a string value

        Parameters
        ----------
        s : str
            The string to parse into a segment
        parser : callable, optional
            A function that transforms a string into a specific type.
            By default, string to int
        """
        res = Segment()
        values = s.replace(LEFT_SEG, "").replace(RIGHT_SEG, "").split(SEPARATOR_SEG)
        for v in values:
            v = v.replace(LEFT_TV, "")
            v = v.replace(RIGHT_TV, "")
            tv = v.split(SEPARATOR_TV)
            res.append(TaggedValue(parser(tv[0]), parse_tag(tv[1])))
        return res


# ------------------------------- #


class LTree(SList):
    """A list of Segment

    Methods
    -------
    init_from_bt(bt, m)
        Create a LTree from a BTree
    init_from_file(filename)
        Initialize a LTree from a file
    write_file(filename)
        Write a file that contains the current instance
    map(kl, kn)
        Applies function to every element of the current instance
    reduce(k, phi, psi_n, psi_l, psi_r)
        Makes a reduction of the current instance into a single value
    uacc(k, phi, psi_n, psi_l, psi_r)
        Processes an upward accumulation into the current instance
    dacc(gl, gr, c, phi_l, phi_r, psi_u, psi_d)
        Processes an downward accumulation into the current instance
    zip(lt)
        Zip the values contained in a second LTree with the ones in the current instance
    map2(lt, f)
        Zip the values contained in a second LTree with the ones in the current instance using a function
    deserialization()
        Get a binary tree from its linear representation
    """

    def __eq__(self, other):
        if isinstance(other, LTree):
            if self.length() != other.length():
                return False
            for i in range(0, self.length()):
                if self[i] != other[i]:
                    return False
            return True
        return False

    def __str__(self):
        res = ""
        for i in range(0, self.length()):
            res = res + str(self[i])
            if i != self.length() - 1:
                res = res + "\n"
        return res

    @staticmethod
    def init_from_bt(bt, m):
        """Create a LTree from a BTree

        Parameters
        ----------
        bt : :obj:`BTree`
            The BTree to transform into a LTree
        m : int
            Variable to define the critical nodes of bt
        """
        def __tv2lv(bt_value):
            val = bt_value.get_value()
            res = Segment()
            res_0 = Segment()
            if bt_value.is_leaf():
                res_0.append(val)
                res.append(res_0)
            else:  # bt_val.is_node()
                res_left = Segment(__tv2lv(bt_value.get_left()))
                res_right = Segment(__tv2lv(bt_value.get_right()))
                if val.is_critical():
                    res_0.append(val)
                    res.append(res_0)
                    res.extend(res_left)
                    res.extend(res_right)
                else:  # val.is_node()
                    res_0.append(val)
                    res_0.extend(res_left[0])
                    res_0.extend(res_right[0])
                    res.append(res_0)
                    res.extend(res_left[1:])
                    res.extend(res_right[1:])
            return res

        # Get a LTree from a BTree
        def up_div(n, m_val): return int(n / m_val) + (0 if n % m_val == 0 else 1)
        bt_one = bt.map(lambda x: 1, lambda x: 1)
        bt_size = bt_one.uacc(lambda x, y, z: x + y + z)
        bt_tags = bt_size.mapt(lambda x: TAG_LEAF,
                               lambda x, y, z: TAG_CRITICAL
                               if up_div(x, m) > up_div(y.get_value(), m) and up_div(x, m) > up_div(z.get_value(), m)
                               else TAG_NODE)
        bt_val = bt.map2(lambda x, y: TaggedValue(x, y), bt_tags)
        return LTree(__tv2lv(bt_val))

    @staticmethod
    def init_from_file(filename, parser=int):
        """Initialize a LTree from a file

        Parameters
        ----------
        filename : str
            The name of the file that contains the LTree to instantiate
        parser : callable, optional
            A function that transforms a string into a specific type.
            By default, string to int
        """
        if filename[-3:] != "." + EXT_FILE_LT:
            filename = filename + "." + EXT_FILE_LT
        res = LTree([])
        with open(filename, "r") as f:
            for line in f:
                a = Segment.from_str(line, parser=parser)
                res.append(a)
        f.close()
        return res

    def write_file(self, filename):
        """Write a file that contains the current instance

        Parameters
        ----------
        filename : str
            The name of the file that we want to write the current instance in
        """
        if filename[-3:] != "." + EXT_FILE_LT:
            filename = filename + "." + EXT_FILE_LT
        with open(filename, "w+") as f:
            f.write(str(self))
        f.close()

    def map(self, kl, kn):
        """Applies function to every element of the current instance

        Precondition
        -------------
        self should empty

        Parameters
        ----------
        kl : callable
            Function to apply to every leaf value of the current instance
        kn : callable
            Function to apply to every node value of the current instance
        """
        assert not self.empty(), "map cannot be applied to an empty linearized tree"
        res = LTree([None] * self.length())
        for i in range(self.length()):
            res[i] = self[i].map_local(kl, kn)
        return res

    def reduce(self, k, phi, psi_n, psi_l, psi_r):
        """Makes a reduction of the current instance into a single value

        The parameters must respect these equalities (closure property):
        * k(l, b, r) = psi_n(l, phi(b), r)
        * psi_n(psi_n(x, l, y), b, r) = psi_n(x, psi_l(l,b,r), y)
        * psi_n(l, b, psi_n(x, r, y)) = psi_n(x, psi_r(l,b,r), y)

        Precondition
        -------------
        self should not be empty

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        phi : callable
            A function used to respect the closure property to allow partial computation
        psi_n : callable
            A function used to respect the closure property to make partial computation
        psi_l : callable
            A function used to respect the closure property to make partial computation on the left
        psi_r : callable
            A function used to respect the closure property to make partial computation on the right
        """
        assert not self.empty(), "reduce cannot be applied to an empty linearized tree"
        tops = Segment([None] * self.length())
        # We start by doing local reductions on each Segment, representing a sub part of the tree
        for i in range(self.length()):
            tops[i] = self[i].reduce_local(k, phi, psi_l, psi_r)
        # The local reductions are reduced into a single value with reduce_global
        return tops.reduce_global(psi_n)

    def uacc(self, k, phi, psi_n, psi_l, psi_r):
        """Processes an upward accumulation into the current instance

        The parameters must respect these equalities (closure property):
        * k(l, b, r) = psi_n(l, phi(b), r)
        * psi_n(psi_n(x, l, y), b, r) = psi_n(x, psi_l(l,b,r), y)
        * psi_n(l, b, psi_n(x, r, y)) = psi_n(x, psi_r(l,b,r), y)

        Precondition
        -------------
        self should not be empty

        Parameters
        ----------
        k : callable
            The function used to reduce a BTree into a single value
        phi : callable
            A function used to respect the closure property to allow partial computation
        psi_n : callable
            A function used to respect the closure property to make partial computation
        psi_l : callable
            A function used to respect the closure property to make partial computation on the left
        psi_r : callable
            A function used to respect the closure property to make partial computation on the right
        """
        assert not self.empty(), "uacc cannot be applied to an empty linearized tree"
        gt = Segment([None]*self.length())
        lt2 = LTree([None]*self.length())
        # We first make a local accumulation to get
        # * Locally non complete accumulated segments
        # * The top value of accumulations, to later pass them for a complete accumulation
        for i in range(self.length()):
            (top, res) = self[i].uacc_local(k, phi, psi_l, psi_r)
            gt[i] = top
            lt2[i] = res

        # We get real top values of accumulation considering a full linearized tree
        gt2 = gt.uacc_global(psi_n)

        res = LTree([None]*gt.length())
        # We update each segment using the real top values calculated previously,
        # the non complete accumulated segments and the initial segments
        # print(gt2)
        # print(lt2)
        for i in range(gt.length()):
            if gt[i].is_node():
                lc = gt2.get_left(i).get_value()
                rc = gt2.get_right(i).get_value()
                res[i] = self[i].uacc_update(lt2[i], k, lc, rc)
            else:
                res[i] = lt2[i]
        return res

    def dacc(self, gl, gr, c, phi_l, phi_r, psi_u, psi_d):
        """Processes an downward accumulation into the current instance

        The parameters must respect these equalities (closure property):
        * gl(c, b) = psi_d(c, phi_l(b))
        * gr(c, b) = psi_d(c, phi_r(b))
        * psi_d(psi_d(c, b), a) = psi_d(c, psi_u(b,a))

        Precondition
        -------------
        self should be empty

        Parameters
        ---------
        gl : callable
            The function used to make an accumulation to the left children in a binary tree
        gr : callable
            The function used to make an accumulation to the right children in a binary tree
        c
            Initial value of accumulation
        phi_l : callable
            A function used to respect the closure property to allow partial computation on the left
        phi_r : callable
            A function used to respect the closure property to allow partial computation on the right
        psi_d : callable
            A function used to respect the closure property to make partial downward accumulation
        psi_u : callable
            A function used to respect the closure property to make partial computation
        """
        assert not self.empty(), "dacc cannot be applied to an empty linearized tree"
        gt = Segment([None] * self.length())
        res = LTree([None] * self.length())
        # We first find the values to pass to sub trees for each segment that contains critical values
        # That is incomplete subtrees (which have node with left and right children not contained in the same segment)
        for i in range(self.length()):
            seg = self[i]
            if seg.has_critical():
                gt[i] = seg.dacc_path(phi_l, phi_r, psi_u)
            else:
                gt[i] = TaggedValue(seg[0].get_value(), "L")

        # We process a global downward accumulation using the initial value of the accumulator
        gt2 = gt.dacc_global(psi_d, c)

        # We finally pass the values of global accumulation to each segment, to update their local accumulation
        for i in range(gt.length()):
            res[i] = self[i].dacc_local(gl, gr, gt2[i].get_value())
        return res

    def zip(self, lt):
        """Zip the values contained in a second LTree with the ones in the current instance

        Precondition
        -------------
        The lengths of self and lt should be equal

        Parameters
        ----------
        lt : :obj:`LTree`
            The LTree to zip with the current instance
        """
        assert self.length() == lt.length(), "The linearized trees have not the same shape"
        res = LTree([None] * self.length())
        for i in range(self.length()):
            res[i] = self[i].zip(lt[i])
        return res

    def map2(self, f, lt):
        """Zip the values contained in a second LTree with the ones in the current instance using a function

        Precondition
        -------------
        The lengths of self and lt should be equal

        Parameters
        ----------
         f : callable
            A function to zip values
        lt : :obj:`LTree`
            The LTree to zip with the current instance
        """
        res = LTree([None] * self.length())
        assert self.length() == lt.length(), "The linearized trees have not the same shape"
        for i in range(self.length()):
            res[i] = self[i].map2(f, lt[i])
        return res

    def deserialization(self):
        """Get a binary tree from its linear representation
        """
        def __graft(bt, lbt, rbt):
            val = bt.get_value()
            if bt.is_node():
                return Node(val, __graft(bt.get_left(), lbt, rbt), __graft(bt.get_right(), lbt, rbt))
            else:  # bt.is_leaf()
                return Node(val, lbt, rbt) if val.is_critical() else bt

        def __remove_annotation(bt):
            v = bt.get_value()
            return Leaf(v.get_value()) if bt.is_leaf() else Node(v.get_value(), __remove_annotation(bt.get_left()),
                                                                 __remove_annotation(bt.get_right()))

        def __lv2ibt(segment):
            stack = []
            has_crit_b = False
            if segment.empty():
                raise EmptyError("An empty Segment cannot be transformed into a BTree")
            for i in range(segment.length() - 1, -1, -1):
                v = segment[i]
                if v.is_leaf():
                    stack.append(Leaf(v))
                elif v.is_critical():
                    stack.append(Leaf(v))
                    if has_crit_b:
                        raise IllFormedError("A ill-formed Segment cannot be transformed into a BTree")
                    else:
                        has_crit_b = True
                else:
                    if len(stack) < 2:
                        raise IllFormedError("A ill-formed Segment cannot be transformed into a BTree")
                    lv = stack.pop()
                    rv = stack.pop()
                    stack.append(Node(v, lv, rv))
            if len(stack) != 1:
                raise IllFormedError("A ill-formed Segment cannot be transformed into a BTree")
            else:
                return has_crit_b, stack[0]

        def __rev_segment_to_trees(lb, glob):
            stack = []
            for i in range(lb.length() - 1, -1, -1):
                if glob[i] == TAG_LEAF:
                    stack.append(lb[i])
                else:  # gt[i] == VTag_Node
                    lbt = stack.pop()
                    rbt = stack.pop()
                    stack.append(__graft(lb[i], lbt, rbt))
            if len(stack) != 1:
                raise IllFormedError("A ill-formed list of incomplete BTree cannot be transformed into a BTree")
            else:
                return stack[0]

        gt = SList()
        list_of_btree = SList()
        for seg in self:
            (has_crit, bt_i) = __lv2ibt(seg)
            list_of_btree.append(bt_i)
            if has_crit:
                gt.append(TAG_NODE)
            else:
                gt.append(TAG_LEAF)
        return __remove_annotation(__rev_segment_to_trees(list_of_btree, gt))
